/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/

/* **************************************************************/
/*								*/
/* Filename:	srvinternals.c					*/
/* Description:	This file contains the implementation of the 	*/
/*		internal functions that are used to communicate	*/
/*		with the AEP Daemon. it includes blocking and	*/
/*		non-blocking command submission functions, and	*/
/*		also commands for querying and waiting on the	*/
/*		completion of commands.				*/
/*								*/
/* Copyright:	Accelerated Encryption Processing Ltd., 2000	*/
/* Author:	Diarmuid O'Neill				*/
/* Product:	AEP1000						*/
/* Date:	13/10/99					*/
/*								*/
/****************************************************************/

#include <stdlib.h>

#include <fcntl.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/uio.h>


#include "aep_i_defs.h"
#include "aep_i_types.h"

#include "aep_api.h"
#include "aep_api_globals.h"
#include "aeptrace.h"
#include "connection.h"
#include "mutex.h"
#include "sockets.h"
#include "aepmalloc.h"
#include "SrvInternals.h"

extern AEP_I_CONNECTION_DATA connection_array[MAX_CONNECTIONS];

/*	AEP_I_Submit:
	This function submits a command to the AEP Daemon without waiting on the reply (Asynchronous).
	The daemon returns a transaction handle for this command which is used to retrieve the reply.

	Inputs	:	Connection Handle  {hConnection}
				Pointer to the Message Block (Header and data) {pCommandBuf}
				Pointer to a Transaction handle (either a valid ptr or NULL)  {phTransHndl}

	Outputs	:	Transaction handle is returned if a valid pointer was passed in phTransHndl
*/
AEP_RV AEP_I_Submit(
        AEP_CONNECTION_HNDL		hConnection,
		AEP_BYTEBLOCK*			pCommandBuf
      )
{
 	AEP_RV rv = AEP_R_OK;

	AEP_MSG_HDR_PTR	message_hdr;
	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_I_Submit", AEP_R_OK, "Entered...");				

 /*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)pCommandBuf->ptr;

  /*Fill out the remaining fields in the message buffer*/
 	message_hdr->app_id = AEP_I_app_ID;

  /*This is an async transaction*/
	message_hdr->flags |= AEP_F_ASYNCMODE;

  /*Write this command to the AEP Daemon*/
	rv = AEP_I_SendToSocket(connection_array[hConnection].descriptor, pCommandBuf->ptr, pCommandBuf->Len);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_Submit", rv, "Error sending Message to the Daemon");				
		return rv;
	}

	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_I_Submit", AEP_R_OK, "...Leaving");				

	return rv; 
}

/*	AEP_I_Transact:
	This function submits a command to the AEP Daemon and waits for the reply (Synchronous).
	
	Inputs	:	Connection Handle  {hConnection}
				Pointer to the Message Block (Header and data) {pCommandBuf}
				Pointer to the Reply Block (Header and data)  {pReplyBlock}

	Outputs	:	The Reply is returned in the reply block.
*/


AEP_RV AEP_I_Transact(
        AEP_CONNECTION_HNDL hConnection,
		AEP_BYTEBLOCK*		pCommandBuf,
		AEP_BYTEBLOCK*		pReplyBuf
      )
{
	AEP_RV rv = AEP_R_OK;

	AEP_MSG_HDR_PTR	message_hdr;
	
	AEP_U8 header[HDR_SIZE];
  

	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_I_Transact", AEP_R_OK, "Entered...");				
	
  /*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)pCommandBuf->ptr;

	/*Fill out the remaining fields in the message buffer*/
	message_hdr->app_id = AEP_I_app_ID;

  /*This is an sync transaction*/
	message_hdr->flags |= AEP_F_SYNCMODE;

	/*Write this command to the	AEP Daemon*/
	rv = AEP_I_SendToSocket(connection_array[hConnection].descriptor, pCommandBuf->ptr, pCommandBuf->Len);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_Transact", rv, "Error sending Message to the Daemon");				
		return rv;
	}

  /*Read the header back from the AEP Daemon*/
	rv = AEP_I_ReadFromSocket(connection_array[hConnection].descriptor, header, HDR_SIZE);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_Transact", rv, "Error reading Reply Header from the Daemon");				
		return rv;
	}

  /*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)header;

  /*Mask off the lower 16 bits of the status field*/
	message_hdr->status &= 0xFFFF0000;
	
  /*Check the status in the header to see how the message was processed.*/
	if (message_hdr->status != AEP_S_COMPLETE)
		return message_hdr->status;

  /*Set the length of the reply block*/
	pReplyBuf->Len = message_hdr->datalen + HDR_SIZE;

  /*Allocate memory for the reply block*/
	pReplyBuf->ptr = AEP_malloc(pReplyBuf->Len);
	
	if(pReplyBuf->ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_Transact", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");				
		return AEP_R_HOST_MEMORY;
	}

  /*Copy the header into the reply block*/	
	memcpy(pReplyBuf->ptr, header, HDR_SIZE);

	if (message_hdr->datalen > 0)
	{
		/*Read the message data from the socket*/
		rv = AEP_I_ReadFromSocket(connection_array[hConnection].descriptor, pReplyBuf->ptr + HDR_DATA_OFFSET, message_hdr->datalen);

		if (rv != AEP_R_OK)
		{
			AEP_free(pReplyBuf->ptr);
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_Transact", rv, "Error : Reading Data from the Daemon");				
			return rv;
		}
	}

	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_I_Transact", AEP_R_OK, "...Leaving");				

	return rv; 
}

/*	AEP_I_Query:
	This function checks to see if a result has become available.
	
	Inputs	:	Connection Handle  {hConnection}
				Transaction Handle {hTransaction}
				Pointer to the Reply Block (Header and data)  {pReplyBlock}

	Outputs	:	The Reply is returned in the reply block.
*/
AEP_RV AEP_I_Query(
        AEP_CONNECTION_HNDL hConnection,
		AEP_BYTEBLOCK*		pCommandBuf,
		AEP_BYTEBLOCK*		pReplyBuf
      )
{
 	AEP_RV rv = AEP_R_OK;

	AEP_MSG_HDR_PTR	message_hdr;
	
	AEP_U8 header[HDR_SIZE];

	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_I_QUERY", AEP_R_OK, "Entered...");				

 /*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)pCommandBuf->ptr;

	/*Fill out the remaining fields in the message buffer*/
	message_hdr->app_id = AEP_I_app_ID;

  /*This is an sync transaction*/
	message_hdr->flags |= AEP_F_ASYNCMODE;

  /*Write this command to the AEP Daemon*/
	rv = AEP_I_SendToSocket(connection_array[hConnection].descriptor, pCommandBuf->ptr, pCommandBuf->Len);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_QUERY", rv, "Error sending message to the Daemon");				
		return rv;
	}
	
  /*Read the header back from the AEP Daemon*/
	rv = AEP_I_ReadFromSocket(connection_array[hConnection].descriptor, header, HDR_SIZE);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_QUERY", rv, "Error reading Header from the Daemon");				
		return rv;
	}

  /*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)header;

  /*Mask off the lower 16 bits of the status field*/
	message_hdr->status &= 0xFFFF0000;
	
  /*Check the status in the header to see how the message was processed.*/
	if (message_hdr->status != AEP_S_COMPLETE)
		return message_hdr->status;


  /*Based on the length indicated by the header just read, realloc the mem and read the data*/
	pReplyBuf->Len = message_hdr->datalen + HDR_SIZE;
	pReplyBuf->ptr = (AEP_U8_PTR) AEP_malloc(pReplyBuf->Len);
		
	if(pReplyBuf->ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_QUERY", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");				
		return AEP_R_HOST_MEMORY;
	}
		
  /*Copy the header into the reply block*/	
	memcpy(pReplyBuf->ptr, header, HDR_SIZE);

	if (message_hdr->datalen > 0)
	{
	  /*Read the data from the AEP Daemon, start into the reply buf after the header info*/
		rv = AEP_I_ReadFromSocket(connection_array[hConnection].descriptor, pReplyBuf->ptr + HDR_DATA_OFFSET, message_hdr->datalen);

		if (rv != AEP_R_OK)
		{
			AEP_free(pReplyBuf->ptr);
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_QUERY", rv, "Error reading Data from the Daemon");							
			return rv;
		}
	}

	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_I_QUERY", AEP_R_OK, "...Leaving");				

	return rv; 
}

/*	AEP_I_Wait:
	This function waits for aresult to become available.
	
	Inputs	:	Connection Handle  {hConnection}
				Transaction Handle {hTransaction}
				Pointer to the Reply Block (Header and data)  {pReplyBlock}

	Outputs	:	The Reply is returned in the reply block.
*/
AEP_RV AEP_I_Wait(
        AEP_CONNECTION_HNDL hConnection,
		AEP_BYTEBLOCK*		pCommandBuf,
		AEP_BYTEBLOCK*		pReplyBuf
      )
{
 	AEP_RV rv = AEP_R_OK;

	AEP_MSG_HDR_PTR	message_hdr;
	
	AEP_U8 header[HDR_SIZE];

	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_I_WAIT", AEP_R_OK, "Entered...");				

 /*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)pCommandBuf->ptr;

	/*Fill out the remaining fields in the message buffer*/
	message_hdr->app_id = AEP_I_app_ID;

  /*This is an sync transaction*/
	message_hdr->flags |= AEP_F_SYNCMODE;

  /*Write this command to the AEP Daemon*/

	rv = AEP_I_SendToSocket(connection_array[hConnection].descriptor, pCommandBuf->ptr, pCommandBuf->Len);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_WAIT", rv, "Error : Sending the Message to the Daemon");				
		return rv;
	}
		
  /*Read the header back from the AEP Daemon*/
	rv = AEP_I_ReadFromSocket(connection_array[hConnection].descriptor, header, HDR_SIZE);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_WAIT", rv, "Error : Reading Header from the Daemon");				
		return rv;
	}

  /*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)header;	

  /*Mask off the lower 16 bits of the status field*/
	message_hdr->status &= 0xFFFF0000;
	
  /*Check the status in the header to see how the message was processed.*/
	if (message_hdr->status != AEP_S_COMPLETE)
		return message_hdr->status;
	
	pReplyBuf->Len = message_hdr->datalen + HDR_SIZE;
		
  /*Based on the length indicated by the header just read, realloc the mem and read the data*/
	pReplyBuf->ptr = (AEP_U8_PTR) AEP_malloc(pReplyBuf->Len);
	if(pReplyBuf->ptr == NULL_PTR)
	{
	
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_WAIT", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");				
		return AEP_R_HOST_MEMORY;
	}
		
  /*Copy the header into the reply block*/	
	memcpy(pReplyBuf->ptr, header, HDR_SIZE);
		
	if (message_hdr->datalen > 0)
	{
		/*Read the message data from the socket*/
		rv = AEP_I_ReadFromSocket(connection_array[hConnection].descriptor, pReplyBuf->ptr + HDR_DATA_OFFSET, message_hdr->datalen);

		if (rv != AEP_R_OK)
		{
			AEP_free(pReplyBuf->ptr);
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_WAIT", rv, "Error : Reading Data from the Daemon");				
			return rv;
		}
	}
	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_I_WAIT", AEP_R_OK, "Leaving...");				

	return rv; 
}


/*	AEP_I_GetDaemonVersion:
	This function gets the version number of the daemon software.
	
	Inputs	:	Connection Handle  {hConnection}
				Pointer to the Command block {pCommandBuf}
				Pointer to the Reply Block (Header and data)  {pReplyBuf}

	Outputs	:	The Reply is returned in the reply block.
*/
AEP_RV AEP_I_GetDaemonVersion(
        AEP_CONNECTION_HNDL hConnection,
	AEP_BYTEBLOCK*		pCommandBuf,
	AEP_BYTEBLOCK*		pReplyBuf )
{
	AEP_RV rv = AEP_R_OK;

	AEP_MSG_HDR_PTR	message_hdr;
	
	AEP_U8 header[HDR_SIZE];

	AEP_TRACE(API, TRACE_DEBUG, FN_NONE, "AEP_I_GetDaemonVersion", AEP_R_OK, "Entered...");				
			
 	/*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)pCommandBuf->ptr;

	/*Fill out the remaining fields in the message buffer*/
	message_hdr->app_id = AEP_I_app_ID;

  	/*This is an sync transaction*/
	message_hdr->flags |= AEP_F_SYNCMODE;

  	/*Write this command to the AEP Daemon*/

	rv = AEP_I_SendToSocket(connection_array[hConnection].descriptor, pCommandBuf->ptr, pCommandBuf->Len);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_GetDaemonVersion", rv, "Error sending Message to the Daemon.");				
			
	return rv;
	}
		
  	/*Read the header back from the AEP Daemon*/
	rv = AEP_I_ReadFromSocket(connection_array[hConnection].descriptor, header, HDR_SIZE);

	if (rv != AEP_R_OK)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_GetDaemonVersion", rv, "Error reading Header from the Daemon");				
		return rv;
	}

  	/*We make this assignemt so we can refer to the message hdr using the msghdr struct*/
	message_hdr = (AEP_MSG_HDR_PTR)header;	

  	/*Mask off the lower 16 bits of the status field*/
	message_hdr->status &= 0xFFFF0000;
	
  	/*Check the status in the header to see how the message was processed.*/
	if (message_hdr->status != AEP_S_COMPLETE)
		return message_hdr->status;
	
	pReplyBuf->Len = message_hdr->datalen + HDR_SIZE;
		
  	/*Based on the length indicated by the header just read, realloc the mem and read the data*/
	pReplyBuf->ptr = (AEP_U8_PTR) AEP_malloc(pReplyBuf->Len);
	if(pReplyBuf->ptr == NULL_PTR)
	{
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_GetDaemonVersion", AEP_R_HOST_MEMORY, "Error: Host Memory Allocation Failure");				
		return AEP_R_HOST_MEMORY;
	}
		
  	/*Copy the header into the reply block*/	
	memcpy(pReplyBuf->ptr, header, HDR_SIZE);
		
	if (message_hdr->datalen > 0)
	{
		rv = AEP_I_ReadFromSocket(connection_array[hConnection].descriptor, pReplyBuf->ptr + HDR_DATA_OFFSET, message_hdr->datalen);
		if (rv != AEP_R_OK)
		{
			AEP_free(pReplyBuf->ptr);
			AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_I_GetDaemonVersion", rv, "Error reading Data from the Daemon");				
			return rv;
		}
	}
	AEP_TRACE(API, TRACE_DEBUG, FN_NONE, "AEP_I_GetDaemonVersion", AEP_R_OK, "Leaving...");				
			
	
	return rv; 
}
