/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/


/* **************************************************************/
/*								*/
/* Filename:	connection.c					*/
/* Description:	This file contains the implementation of the 	*/
/*		connection management functions. It includes 	*/
/*		functions to open and close connections, and 	*/
/*		a function to get connection data from the 	*/
/*		connection hash table.				*/
/*								*/
/* Copyright:	Accelerated Encryption Processing Ltd., 2000	*/
/* Author:	Diarmuid O'Neill				*/
/* Product:	AEP1000						*/
/* Date:	12/10/99					*/
/*								*/
/****************************************************************/



#include <stdlib.h>
#include <string.h>

#include <stdio.h>
#include "aep_lib.h"

#include "aep_i_defs.h"
#include "aep_i_types.h"
#include "aep_api.h"
#include "aep_i_api.h"
#include "aep_api_globals.h"
#include "aeptrace.h"
#include "sockets.h"
#include "mutex.h"
#include "aepmalloc.h"
#include "connection.h"

#ifdef AEP_Win32
#include <io.h>
#endif

static AEP_RV AEP_I_GetConnHndl(AEP_CONNECTION_HNDL *hConnection);

static AEP_RV AEP_I_FreeConnHndl(AEP_CONNECTION_HNDL hConnection);


/*Array of connection data*/
AEP_I_CONNECTION_DATA connection_array[MAX_CONNECTIONS];

static pthread_mutex_t AEPAPI_connection_mutex = PTHREAD_MUTEX_INITIALIZER;

/*	AEP_OpenConnection:
	This function opens a connection to the AEP Daemon and returns a connection
	handle.  It stores the connection data in the connection hash table.  

	Inputs	: None
	Outputs	: Pointer to a Connection Handle {phConnection}
*/
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_OpenConnection(
        AEP_CONNECTION_HNDL*	phConnection
)
{
	AEP_RV rv = AEP_R_OK;									/*Return value*/

	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_OpenConnection", AEP_R_OK, "Entered...");

/**Lock the flags mutex as we are about to read it.  Note since this is the only place that
	new connection handles are generated, I also use this mutex to protect the handle generation*/
	_LOCK(&AEPAPI_flags_mutex);

  /* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED))
	{
		_UNLOCK(&AEPAPI_flags_mutex);
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_OpenConnection", AEP_R_AEPAPI_NOT_INITIALIZED, "Error : API not Initialised");
		return AEP_R_AEPAPI_NOT_INITIALIZED;
	}
 
/**Unlock the flags mutex*/
	_UNLOCK(&AEPAPI_flags_mutex);

  /* Get the next free connection handle */
	rv= AEP_I_GetConnHndl(phConnection);
	if(rv != AEP_R_OK)
   	{
		_UNLOCK(&AEPAPI_flags_mutex);
		/*AEP_free(new_connection);*/
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_OpenConnection", rv, "Error : Generating new Connection Handle");

		return rv; 
    	}
  
	/*Open the connection to the AEP daemon*/
	rv = AEP_I_DaemonConnect(&(connection_array[*phConnection].descriptor));
	if(rv != AEP_R_OK)
    {
		/*AEP_free(new_connection);
		_UNLOCK(&AEPAPI_hashtable_mutex);*/
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_OpenConnection", rv, "Error : Connecting to Daemon");

		return rv; 
    }

	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_OpenConnection", AEP_R_OK, "for Connection %lu... complete", *phConnection);
 
	return rv;
}
/* End AEP_OpenConnection*/


/*	AEP_CloseConnection:
	This function closes a connection to the AEP Daemon given the connection
	handle.  It removes the connection data from the connection has table.

	Inputs	: Connection Handle.  {hConnection}
	Outputs	: None
*/
AEP_SLIB_CALLING_CONVENTION AEP_RV AEP_CloseConnection(
        AEP_CONNECTION_HNDL		hConnection
      )
{
	AEP_RV rv = AEP_R_OK;

	AEP_TRACE(API, TRACE_DEBUG, FN_IN, "AEP_CloseConnection", AEP_R_OK, "Entered.... Connection %lu", hConnection);
  
/* Protect the global flags variable */ 
	_LOCK(&AEPAPI_flags_mutex);
  /* make sure we are initialized */
	if (!(AEP_I_global_flags & AEP_IGF_INITIALIZED)) 
    {
		_UNLOCK(&AEPAPI_flags_mutex);
		AEP_TRACE(API, TRACE_ERRORS, FN_NONE, "AEP_CloseConnection", AEP_R_AEPAPI_NOT_INITIALIZED, "Error : API not Initialised");		

		return AEP_R_AEPAPI_NOT_INITIALIZED;
    }
/*Finished with the global flags variable */
	_UNLOCK(&AEPAPI_flags_mutex);

	AEP_I_DaemonDisconnect(connection_array[hConnection].descriptor);
	
	AEP_I_FreeConnHndl(hConnection);

	AEP_TRACE(API, TRACE_DEBUG, FN_OUT, "AEP_CloseConnection", AEP_R_OK, "...Leaving");

	return rv; 
}
/*End AEP_CloseConnection*/


/*Used to make a guess at where the next free handle is*/
static AEP_U32 next_free_hndl_maybe = 1;

/*Used to determine if we've any free handles left*/
static AEP_U32 num_used_conn_hndls = 0;

/*	AEP_I_GetConnHndl:
	This internal function gets the next free connection handle

	Inputs	: Ptr to Connection Handle  {hConnection}
	Outputs	: 
*/
static AEP_RV AEP_I_GetConnHndl(
  AEP_CONNECTION_HNDL			*phConnection
)
{
	AEP_RV rv = AEP_R_OK;
	AEP_U32 index;
	
/*Lock the connection array mutex*/
	_LOCK(&AEPAPI_connection_mutex);
	
  /*Are there any handles left?*/
	if (num_used_conn_hndls >= MAX_CONNECTIONS)
	{
		_UNLOCK(&AEPAPI_connection_mutex);
		
		return AEP_R_NO_MORE_CONNECTION_HNDLS;
	}

  /*Guess the next free handle's index*/
	index = next_free_hndl_maybe;

  /*Find a free handle*/	
	while (connection_array[index].status != 0)
	{
		index++;
		if(index >= MAX_CONNECTIONS)	
			index = 0;
	}

  /*Sanity check, as if there are handles left we should never enter here*/  
  	if (connection_array[index].status != 0)
	{
/*UnLock the connection array mutex*/
		_UNLOCK(&AEPAPI_connection_mutex);
		return AEP_R_NO_MORE_CONNECTION_HNDLS;
	}
	
  /*Mark the handle as used*/	
	connection_array[index].status = 1;
 
  /*Ready the return value*/ 
	*phConnection = index;

  /*This handle certainly is not free any more, maybe the next one is*/	
	next_free_hndl_maybe++;
	
  /*See if this must wrap*/
	if (next_free_hndl_maybe >= MAX_CONNECTIONS)
		next_free_hndl_maybe = 0;
		
/*UnLock the connection array mutex*/
	_UNLOCK(&AEPAPI_connection_mutex);
	  
	return rv;
}
/*End AEP_I_FreeConnectionData*/

/*	AEP_I_GetConnHndl:
	This internal function gets the next free connection handle

	Inputs	: Ptr to Connection Handle  {hConnection}
	Outputs	: 
*/
static AEP_RV AEP_I_FreeConnHndl(
  AEP_CONNECTION_HNDL			hConnection
)
{
	AEP_RV rv = AEP_R_OK;

/*Lock the connection array mutex*/
	_LOCK(&AEPAPI_connection_mutex);

  /*Mark the handle as free*/	
	connection_array[hConnection].status = 0;

	next_free_hndl_maybe = hConnection; 
  
/*UnLock the connection array mutex*/
	_UNLOCK(&AEPAPI_connection_mutex);
	return rv;
}
/*End AEP_I_GetConnectionData*/

/* handle generation functions */
AEP_U32 AEP_I_tran_id_counter; /* must not start with 0 'cause it's AEP_INVALID_HANDLE */

AEP_RV AEP_I_NewTranID(
  AEP_U32_PTR id                 /* handle returned */
)
{
  if(AEP_I_tran_id_counter == I_MAX_HANDLE)
	AEP_I_tran_id_counter = 1;
    
  *id = (AEP_I_tran_id_counter++);

  return AEP_R_OK;
}
