/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/



#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <string.h>
#include "aep_i_defs.h"
#include "aep_i_types.h"
#include "config.h"
#include "config_hndlr.h"
#include "hash.h"
#include "mutex.h"
#include "trace.h"
#include "sockets.h"
#include "aependian.h"
#include "keyman_hndlr.h"
#include "buf_hndlr.h"
#include "response_hndlr.h"
#include "ctx_hndlr.h"
#include "timeout_hndlr.h"
#include "aepmalloc.h"
#include "aeptrace.h"
#include "aepdaemon.h"
#include "daemon_version.h"
#include "aep_version.h"


/*
AEP Daemon process 7th November Diarmuid O'Neill

This is the main module of the aep daemon process.  This process
is essentially a TCP/IP server program that accepts connections from
applications using the AEP API and organises the messages sent accross
these connections for processing by the AEP accelerator
*/


/***************************************************
	main:
	This is the entry point to the aepdaemon program

	Inputs	:	None

	Outputs	:	none
*****************************************************/

#ifndef STDOUT_FILENO
	#define STDOUT_FILENO 1
#endif


#define ISDAEMON 1
int main()
{
	struct rlimit rlim;   /* Used to increase the no. of file descriptors available to this process*/
	int sock_fd, new_sock;  /* listen on sock_fd, new connection on new_sock */
	
	int i; /*counter*/
	struct sigaction act;

	pid_t pid;
	int fd;	
	
	AEP_U32 rv = AEP_R_OK;

	/*Used in the creation of new threads*/
	pthread_attr_t attr1,attr2;
	size_t stacksize;

	pthread_t pthrMain; /*Main thread*/
  	pthread_t pthrConnHndlr; /*Thread to read requests from the na */
	pthread_t pthrASICBuffHndlr; /*This thread handles the ASIC buffer emptying*/
	pthread_t pthrSABuffHndlr; /*This thread handles the SA buffer emptying*/
	pthread_t pthrResponseHndlr;  /*This thread handles the reading of responses*/
	pthread_t pthrTimedOutHndlr;  /*This thread handles the disposal of timed out requests*/
	pthread_t pthrKeyManHndlr;

/*      pthread_t pthrDESBuffHndlr; This thread handles the DES buffer emptying*/
/*      pthread_t pthrContextHndlr; Thread to handle contexts */
	
  /*Insert log entry*/
  AEP_TRACE(DAEMON, TRACE_DEBUG, FN_IN, "AEP_Daemon", AEP_R_OK, "Starting...");
	

#ifdef ISDAEMON

  /*Start the daemon creation process*/
	if (fork() != 0)
		exit(0);
	else
	{
		/*Child becomes the daemon*/
		/*Create new session with this process as the leader*/
		pid = setsid();
		
		/*Close stdout*/
		close(STDOUT_FILENO);
		
		/*Open the console which will now become stdout*/	
		fd = open("/dev/console", O_WRONLY|O_NOCTTY);
	}
#endif



  /*Ignore broken pipe signals in case we read from a closed socket.  The default action if
	we get a SIGPIPE signal is to die.  If we ignore this sig then reads and writes to closed
	sockets will result in a error returned rather than program death. (which was nice)*/
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
		perror("Can't ignore SIG_PIPE signals\n");



	/* At start up need to load values from aep.conf */	
	rv = PresetConfigValues();
	if (rv != AEP_R_OK)
	{
	  AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", rv, "Error : Could not Preset Configuration Variables - Exiting...");
		
		perror("Error configuring variables\n");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Daemon Configuration Complete");

	/*Catch SIGHUP signal that the management application sends us this tell us that
	the configuration variables have changed and we should reload them */

	act.sa_handler = SetConfigValues;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	
	if (sigaction(SIGHUP, &act, NULL) < 0)
	{
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error : Can`t catch SIGHUP signals");
		perror("Can't catch SIGHUP signals\n");
	}


	/*Open the accelerator PCI device */
	AccFd = open(AEP_DeviceName,O_RDWR);

	if (AccFd < 0)
	{
		AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_TARGET_ERROR, "Error : Could not open Accelerator Device");
		perror("Error communicating with the accelerator");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Successful Connection to Accelerator Device");


	
  /*Set the applications process Id (Used to make the unique transaction hndl)*/
	proc_ID = getpid();

  /*Initialise the response hashtable*/
	rv = AEP_I_InitHashtable(&AEP_I_response_table.response_table, ResponseTableSize);
	if(rv != AEP_R_OK)
	{
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", rv, "Error setting up the Hash Table");
		perror("Error setting up internal structures : Hash Table");
		exit(1);
	}

  /*Change the limititation on the number of open file descriptors to the max number of
	open file descriptors for a process*/
	if(getrlimit(RLIMIT_NOFILE,&rlim))
	{
		AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error increasing File Descriptors");
		perror("Error reading the allowed number of file descriptors per process\n");
	}
	else
	{
	  /*Set current limit equal to the maximum limit*/
		rlim.rlim_cur=rlim.rlim_max;
		if(setrlimit(RLIMIT_NOFILE,&rlim))
		{
			AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error increasing File Descriptors");
			perror("Error setting the allowed number of file descriptors per process\n");
		}
	}

  /*Initialise the response item mutexes*/
  	for (i=0;i < ResponseTableSize;i++)
	{
		if (pthread_mutex_init(&response_item_mutex_array[i], NULL) != 0)
		{
		     AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_MUTEX_BAD, "Error setting up Response Mutex Array - Exiting...");
			perror("Error setting up response mutex array");
			exit(1);
		}	
	}

  /*Initialise the semaphore used to limit the amount of open connections
  Limit the number of connections to the max number of open fds - 10*/
	SEM_INIT(&connection_counter, 0, (int)rlim.rlim_cur-10);
	
  /*Initialise the semaphore used to wait for authentication*/
	SEM_INIT(&auth_semaphore, 0, 0);
	
  /*Initialise the semaphore used to wait for the read done array signal*/
	SEM_INIT(&read_done_array_semaphore, 0, 0);
	
  /*Get main threads id*/
	pthrMain = pthread_self();

  /*printf("AEP Daemon process starting...\n"); */

  /*Set up the daemon socket and start it listening for connections*/
	rv=AEP_I_DaemonListen(&sock_fd);
	if(rv != AEP_R_OK)
	{
		AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_SOCKERROR, "Error setting up the Daemon Socket - Exiting...");
		perror("Error setting up internal structures daemon socket");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Daemon Listening for Connections");

  /*At the moment i'm creating detached threads i.e. main does not have to wait for the threads*/
	pthread_attr_init(&attr1);
	pthread_attr_setdetachstate(&attr1, PTHREAD_CREATE_DETACHED);
	pthread_attr_setscope(&attr1, PTHREAD_SCOPE_SYSTEM);
			
  /*Try to make a connection to the key db*/
  	rv = AEP_I_KeyDBConnect(&keydb_fd, KEY_DB_PORT);
	
	if(rv != AEP_R_OK)
	{
		keydb_fd = -1;
	/*	printf("No key db\n");   */
	}
	else
	{
	  /*Create the Key management handler thread*/
		if (pthread_create(&pthrKeyManHndlr, &attr1, ptrKeyManHndlr, NULL))
		{
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error setting up Key Management Handler Thread - Exiting...");
			perror("Error setting up internal structures\n");
			exit(1);
		}
		else
			AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Key Management Handler Thread - OK");
	
	}

  /*Create the ASIC buffer emptying thread*/
	if (pthread_create(&pthrASICBuffHndlr, &attr1, ptrASICBuffHndlr, NULL))
	{
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error setting up ASIC Buffer Handler Thread - Exiting...");
		perror("Error setting up internal structures\n");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "ASIC Buffer Thread - OK");
  
  /*Create the DES buffer emptying thread*/
/*	if (pthread_create(&pthrDESBuffHndlr, &attr1, ptrDESBuffHndlr, NULL))
	{
		AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error setting up DES Buffer Handler Thread - Exiting...");
		perror("Error setting up internal structures\n");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "DES Buffer Thread - OK");
	
	*/
	 
  /*Create the SA buffer emptying thread*/
	if (pthread_create(&pthrSABuffHndlr, &attr1, ptrSABuffHndlr, NULL))
	{
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error setting up Strong ARM Buffer Handler Thread - Exiting...");
		perror("Error setting up internal structures\n");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "SA Buffer Thread - OK");

	/* NOTE : On HP-UX the default stack size for the response handler thread is 
	too small. In this section we set the stack size at 1M if less than that. */

	/* Set the size of the stack for the response thread if it is less than 1M bytes */ 	
	pthread_attr_init(&attr2);
	pthread_attr_getstacksize(&attr2, &stacksize);
	 	
	if (stacksize < (1024*1024))
	{
		pthread_attr_setstacksize(&attr2, 1024*1024); 
	}
	
	pthread_attr_getstacksize(&attr2, &stacksize);
  	
	/*Create the response handler thread*/
	if (pthread_create(&pthrResponseHndlr, &attr2, ptrResponseHndlr, NULL))
	{
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error setting up Response Buffer Handler Thread - Exiting...");
		perror("Error setting up internal structures\n");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Response Handler Thread - OK");

    /*Create the timed out Handler thread*/
	if (pthread_create(&pthrTimedOutHndlr, &attr1, ptrTimedOutHndlr, NULL))
	{
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error setting up Timed Out Handler Thread - Exiting...");
		perror("Error setting up internal structures\n");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Timeout Handler Thread - OK");

#ifdef AEP_CTX_ON

  /*Create the context handler thread*/
	if (pthread_create(&pthrContextHndlr, &attr1, ptrContextHndlr, NULL))
	{
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error setting up Context Handler Thread - Exiting...");
		perror("Error setting up internal structures\n");
		exit(1);
	}
	else
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Context Handler Thread - OK");

  /*Wait for the context handler thread to signal us before we continue*/
	SEM_WAIT(&auth_semaphore);

#endif

  /*printf("AEP Daemon process waiting connections...\n");*/
	
	while(1)
    {	

	  /*This decrements the semaphore and if it's zero blocks, so we can create threads
		until this value becomes zero.  When the threads exit the semaphore is incremented*/
		SEM_WAIT(&connection_counter);
		
	/*	printf("About to wait on the connection\n");*/
		
	  /*Wait for a connection to arrive, and return the new socket descriptor*/
		rv=AEP_I_DaemonWait(&sock_fd, &new_sock);
	  	if(rv != AEP_R_OK)
		{
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_SOCKERROR, "Error setting up the Daemon Socket - Exiting...");
			perror("Error setting while accepting connection");
			exit(1);
		}

	  /*Insert log entry*/
  	    AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Connection Received");
		
      /*Create the connection handler thread*/
		if (pthread_create(&pthrConnHndlr, &attr1, ptrConnHndlr, (void *)new_sock))
      	{
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_GENERAL_ERROR, "Error : Spawning a Thread to handle a Connection");
			close (new_sock);
			perror("Error spawning a thread to handle a connection\n");
		}

    }

	/*Exit the main thread*/
/*	pthread_exit((void *)main_thr);
	close(AccFd);*/
	return 0;
}

/*	ptrConnHndlr
	This function defines the thread that handles the connection

	Inputs	:	Args to thread which is the socket descriptor that refers to the 
				new connection.  {arg}
				
	Outputs	:	None
*/
static void *ptrConnHndlr(void *arg)
{
  /*Socket descriptor initialised from the arg passed to the thread*/
	int sock_fd = (int)arg;

  /*Return value*/
	AEP_U32 rv = AEP_R_OK;

  /*These two variables are used in changing this threads priority*/
	struct sched_param param;
	int policy;

  /*Pointer to a message header structure*/
	AEP_MSG_HDR_PTR	message_hdr;

/*	printf("got a con\n");*/

  /*Insert log entry*/
    AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "ptrConnHndlr", AEP_R_OK, "Entered...");
    AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "ptrConnHndlr", AEP_R_OK, "Connection Handler Thread started");

  /*Make this thread a lower priority.  Note : maybe we do not need to do this
	for every connection maybe should just change the inherited priority
	and then change the priority of the other threads higher thus not changing pri
	for every connection opened*/
	rv = pthread_getschedparam (pthread_self(), &policy, &param);
	param.sched_priority = param.sched_priority + 1;
	rv = pthread_setschedparam (pthread_self(), policy, &param);

	while(1)
	{
  		message_hdr = (AEP_MSG_HDR_PTR) AEP_malloc(sizeof(AEP_MSG_HDR));

		if (message_hdr == NULL)
		{
			close(sock_fd);
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "ptrConnHndlr", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure. Exiting Thread");

		  /*Increment the semaphore allowing another possible connection*/
			sem_post(&connection_counter);

			pthread_exit(NULL);
		}

		/*Read the message header from the client*/
		rv = AEP_I_ReadFromSocket(sock_fd, (AEP_U8*)message_hdr, HDR_SIZE);

		/*Note need to change this here to see if the socket has been closed.
		ie. returned 0 so then we just close the sock and exit the thread but there
		is no need to exit the app*/
		if (rv == AEP_R_SOCKEOF)
		{
			close(sock_fd);
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "ptrConnHndlr", rv, "Client closed Connection. Exiting Thread");
			
/*			printf("client closed\n");*/

		  /*Increment the semaphore allowing another possible connection*/
			sem_post(&connection_counter);

			AEP_free (message_hdr);

			pthread_exit(NULL);
		}
		else if(rv != AEP_R_OK)
		{
			close(sock_fd);
			perror("Error reading from the socket");
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "ptrConnHndlr", rv, "Error reading from the Socket. Exiting Thread");

		  /*Increment the semaphore allowing another possible connection*/
			sem_post(&connection_counter);

			AEP_free (message_hdr);

			pthread_exit(NULL);
		}

		/*Insert log entry*/
	    AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Request received from App %d", message_hdr->app_id);

		/*Depending on the command type call one of the command handling functions*/
		switch(message_hdr->flags & AEP_F_CMDMASK)
		{
		case AEP_F_SRVCMD:
  			rv = DoDaemonCommand(message_hdr, sock_fd);
			break;
		
		case AEP_F_NOCTXCMD:
			rv = DoDeviceCommand(message_hdr, sock_fd);
			break;
			
		case AEP_F_CTXCMD:
#ifdef AEP_CTX_ON
			rv = DoContextCommand(message_hdr, sock_fd);
			break;
#else
			break;
#endif	
			
		case AEP_F_KEYMANCMD:
			rv = DoKeyManCommand(message_hdr, sock_fd);
			break;
		
		default:
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_SRV", AEP_R_GENERAL_ERROR, "Error : Invalid Message received and discarded");
			break;
		}
		
		if(rv != AEP_R_OK)
		{
		/*	printf("Error processing command\n");*/
		     AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "ptrConnHndlr", rv, "Error : Error processing Command. Exiting Thread");

			close(sock_fd);

			/*Increment the semaphore allowing another possible connection*/
			sem_post(&connection_counter);

			pthread_exit(NULL);
		}
	}

	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_OUT, "ptrConnHndlr", AEP_R_OK, "Leaving...");
	return ((void *)NULL);
}

/*	DoDaemonCommand
	This function handles daemon commands.  E.g. query/wait etc.

	Inputs	:	Request header.  {message_hdr}
				Socket Descriptor of the connection the command is comming over. {sock_fd}
				
	Outputs	:	Returns an AEP_RV, indicating success or failure to the calling function.
*/
AEP_RV DoDaemonCommand(AEP_MSG_HDR_PTR message_hdr, int sock_fd)
{
  /*Return value*/
	AEP_U32 rv = AEP_R_OK;
	
	AEP_U32 hash_value;
	
	AEP_DAEMON_VERSION daemonVersion;

  /*Pointer to response hash table element*/
	AEP_I_RESPONSE_DATA_PTR pResponse;

	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "DoDaemonCommand", AEP_R_OK, "Entered...");

  /*Get the hash value for this transaction*/
	hash_value = AEP_I_HASH(message_hdr->trans_hndl);

/*Lock access to the response table*/

	_LOCK(&response_item_mutex_array[hash_value]);

  /*Depending on the daemon command being executed, handle appropriately*/
	switch (message_hdr->command)
	{
	  /*Command is query i.e. check if the response for the given trans hndl is available*/
		case AEP_C_QUERY:
			
		  /*Insert log entry*/
	  	    AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Handling a Query command");

		  /*Get the entry in the hash table for this trans hndl*/
			rv = AEP_I_HashGetEntry(AEP_I_response_table.response_table,message_hdr->trans_hndl,
			(AEP_VOID_PTR)&pResponse);

		  /*If it's not there, this is an invalid transaction handle*/
			if(rv != AEP_R_OK)
			{	
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : No Entry for this Transaction Handle");
				
				message_hdr->status = AEP_S_UNKNOWN;

			  /*Send the message header back to the app indicating status*/
				rv = AEP_I_SendToSocket(sock_fd, (AEP_U8*)message_hdr, HDR_SIZE);
				if (rv != AEP_R_OK)
				{
					AEP_free (message_hdr);
					_UNLOCK(&response_item_mutex_array[hash_value]);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API");
					return rv;
				}
			}
		  /*If it's not ready yet return this info*/
			else if(((AEP_MSG_HDR_PTR)pResponse->msg)->status < AEP_S_COMPLETE)
			{
			    AEP_TRACE(DAEMON, TRACE_WARNINGS, FN_NONE, "AEP_Daemon", AEP_R_TRANSACTION_NOT_READY, "Transaction not yet ready");
				
				/*Send the response entry's header back to the app indicating status*/
				rv = AEP_I_SendToSocket(sock_fd, (AEP_U8*)pResponse->msg, HDR_SIZE);
				if (rv != AEP_R_OK)
				{
					AEP_free (message_hdr);
					_UNLOCK(&response_item_mutex_array[hash_value]);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API");
					return rv;
				}
			}
			else if(pResponse->response_semaphore != NULL)
			{
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_Daemon", AEP_R_TRANSACTION_CLAIMED, "Thread already waiting on this Response");

				message_hdr->status = AEP_S_CLAIMED;

			  /*Send the response entry's header back to the app indicating status*/
				rv = AEP_I_SendToSocket(sock_fd, (AEP_U8*)message_hdr, HDR_SIZE);
				if (rv != AEP_R_OK)
				{
					AEP_free (message_hdr);
					_UNLOCK(&response_item_mutex_array[hash_value]);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API");
					return rv;
				}
			}
		  /*Otherwise it's ready in the response table so pack it up and ship it out*/
			else
			{
			  /*Send this now complete response back to the application*/
				rv = AEP_I_SendToSocket(sock_fd, (AEP_U8*)pResponse->msg, 
					((AEP_MSG_HDR_PTR)pResponse->msg)->datalen + HDR_SIZE);
			
				if (rv != AEP_R_OK)
				{
				  /*Remove the response entry from the hash table*/
					AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
				
					AEP_free (message_hdr);
					AEP_free(pResponse->msg);
				  	AEP_free(pResponse);

					_UNLOCK(&response_item_mutex_array[hash_value]);

				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API");		
					return rv;
				}
				
			  /*Remove the entry from the hash table*/
				rv = AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
				if(rv != AEP_R_OK)
				{
					AEP_free (message_hdr);
					AEP_free(pResponse->msg);
				  	AEP_free(pResponse);
									
					_UNLOCK(&response_item_mutex_array[hash_value]);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error removing response from the Hash Table");
					return rv; 
				}
				AEP_free (pResponse->msg);
				AEP_free (pResponse);
			}
			/*Unlock the response table mutex*/
			_UNLOCK(&response_item_mutex_array[hash_value]);
		break;
	  /*Command is wait i.e. check if the response for the given trans hndl is available, if not wait*/		
		case AEP_C_WAIT:
		
		  /*Insert log entry*/
	  	    AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Handling a Wait Command");
			
		  /*Get the entry in the hash table for this trans hndl*/
			rv = AEP_I_HashGetEntry(AEP_I_response_table.response_table,message_hdr->trans_hndl,
					(AEP_VOID_PTR)&pResponse);

		  /*If it's not there, this is an invalid transaction handle*/
			if(rv != AEP_R_OK)
			{	
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "No Entry for this Transaction Handle");
				
				message_hdr->status = AEP_S_UNKNOWN;
				message_hdr->datalen = 0;
		
			  /*Send the status info back to the app*/
				rv = AEP_I_SendToSocket(sock_fd, (AEP_U8*)message_hdr, HDR_SIZE);
				if (rv != AEP_R_OK)
				{
					AEP_free (message_hdr);
					_UNLOCK(&response_item_mutex_array[hash_value]);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API");
					return rv;
				}
			}
		  /*There's already a thread waiting for this response to become available*/
			else if(pResponse->response_semaphore != NULL)
			{
			    AEP_TRACE(DAEMON, TRACE_WARNINGS, FN_NONE, "DoDaemonCommand", AEP_R_TRANSACTION_CLAIMED, "Thread already waiting on this Response");

				message_hdr->status = AEP_S_CLAIMED;

				rv = AEP_I_SendToSocket(sock_fd, (AEP_U8*)message_hdr, HDR_SIZE);
				if (rv != AEP_R_OK)
				{
					AEP_free (message_hdr);
					_UNLOCK(&response_item_mutex_array[hash_value]);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API");
					return rv;
				}
			}
			else
			{
			  /*If the response has not yet returned from the accelerator*/
				if(((AEP_MSG_HDR_PTR)pResponse->msg)->status < AEP_S_COMPLETE)
				{
					/*Record the socket descriptor number*/
					pResponse->sock_fd = sock_fd;
					
					pResponse->response_semaphore = (sem_t *) AEP_malloc(sizeof(sem_t));
					if (pResponse->response_semaphore == NULL)
					{
						AEP_free (message_hdr);
						_UNLOCK(&response_item_mutex_array[hash_value]);
					    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");
						return AEP_R_HOST_MEMORY;
					}

					SEM_INIT(pResponse->response_semaphore, 0, 0);

					/*Does not matter that response may have arrived between here and the if
					above as the post will add one to the semaphore and we will not wait at all*/
					_UNLOCK(&response_item_mutex_array[hash_value]);
	
					SEM_WAIT(pResponse->response_semaphore);
					
					sem_destroy(pResponse->response_semaphore);
					AEP_free(pResponse->response_semaphore);
					
					pResponse->response_semaphore = NULL;
				}
				else
				{
				  /*Send the completed reply back to the application*/
					rv = AEP_I_SendToSocket(sock_fd, (AEP_U8*)pResponse->msg, 
						((AEP_MSG_HDR_PTR)pResponse->msg)->datalen + HDR_SIZE);
					if (rv != AEP_R_OK)
					{			
					  /*Remove the response entry from the hash table*/
						AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
				
						AEP_free (message_hdr);
				
						AEP_free(pResponse->msg);
					  	AEP_free(pResponse);
					
						_UNLOCK(&response_item_mutex_array[hash_value]);

					    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API");

						return rv;
					}
			
				  /*Remove the entry from the responses table*/
					rv = AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
					if(rv != AEP_R_OK)
					{
						AEP_free (message_hdr);
						AEP_free (pResponse->msg);
						AEP_free(pResponse);
					
						_UNLOCK(&response_item_mutex_array[hash_value]);
				
					    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error removing Response from Hash Table");						
						
						return rv; 
					}
					
					/*Free the response memory*/
					AEP_free (pResponse->msg);
					
					/*Unlock the response table mutex*/
					_UNLOCK(&response_item_mutex_array[hash_value]);
				}
						
			  /*Free the mem used to hold the response*/
				AEP_free (pResponse);
			}
		break;
		
		case AEP_C_GET_DAEMON_VERSION:
	
			/*Return the version number of the daemon software*/
	
			_UNLOCK(&response_item_mutex_array[hash_value]);
			
			 /*Insert log entry*/
			 AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Handling a Get Daemon Version Command.");						
						
			
			/*Set the message status to complete */
			message_hdr->status = AEP_S_COMPLETE;
			message_hdr->datalen = sizeof(AEP_DAEMON_VERSION);
			
			/*Send the header back to the application*/
			rv = AEP_I_SendToSocket( sock_fd, (AEP_U8*)message_hdr, HDR_SIZE);
			if (rv != AEP_R_OK)
			{							
				
				AEP_free (message_hdr);
				
				 AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API.");						
			
				return rv;
			}

			
			/*Fill the daemon version number struct with the appropriate values*/
			/*These values are #defined in daemon_version.h*/
			
			strcpy(daemonVersion.versionString, DAEMON_VERSION);
			strcat(daemonVersion.versionString, COMPILER_VERSION);
			
			/*Now send back the response, i.e. the daemon version number*/
			rv = AEP_I_SendToSocket( sock_fd, (AEP_U8*) &daemonVersion, sizeof(AEP_DAEMON_VERSION));
			if (rv != AEP_R_OK)
			{				
				AEP_free (message_hdr);
				
				AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDaemonCommand", rv, "Error : Writing Data to the API.");						
			
				return rv;
			}
			

			AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "Retreived Daemon Version");						
			
			break;
	}

  /*Free the mem used to hold the request from the app*/
	AEP_free (message_hdr);
	
	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_OUT, "DoDaemonCommand", AEP_R_OK, "Leaving...");
	return rv;
}

/*	DoDeviceCommand
	This function handles accelerator commands.

	Inputs	:	Request header.  {message_hdr}
				Socket Descriptor of the connection the command is comming over. {sock_fd}
				
	Outputs	:	Returns an AEP_RV, indicating success or failure to the calling function.
*/
AEP_RV DoDeviceCommand(AEP_MSG_HDR_PTR message_hdr, int sock_fd)
{
  /*Return value*/
	AEP_U32 rv = AEP_R_OK;

  /*Pointer to response hash table element*/
	AEP_I_RESPONSE_DATA_PTR pResponse;
	
	AEP_U32 hash_value;

  /*The device command will be added to one of the OutGoing Request Buffers.  Rather than
	write seperate code for each of the buffers we simply check at the start what buffer 
	the request should be added to and then get pointers to the necessary variables and use
	common code to do the work.  The following data structures are used to hold these ptrs*/

  /*Used to point to the RequestBuffer Mutex for this request*/
	pthread_mutex_t *AEP_RequestBuffer_mutex;

  /*Used to point to the condition variables for the sending/waiting to be sent of this buffer*/
	pthread_cond_t *send_now, *send_complete;

  /*Used to point to the offset into the request Buffer for this request*/
	AEP_U32 *BuffIndex;

  /*Used to point to the variable used when deciding if this buffer should go to the accelerator*/
	AEP_U32 *SendBuffer;

  /*Used to point to the Request Buffer for this request*/
	AEP_U64 **AEP_RequestBuffer;

  /*Used to hold the size of the Request buffer associated with this request*/
	AEP_U32 WRITE_BUFFER_SIZE;

  /*Used to determine if and how much padding is necessary*/
	AEP_U32 padding, msg_len;

  /*Insert log entry*/
	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "DoDeviceCommand", AEP_R_OK, "Entered...");
    AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "DoDeviceCommand", AEP_R_OK, "Handling a Command. Type %d", message_hdr->command);

  /*Get the hash value for this transaction*/
	hash_value = AEP_I_HASH(message_hdr->trans_hndl);

  

  /*Determine which request buffer this request is going to be added to and set
  the variables up appropriately*/
	switch(message_hdr->flags & AEP_F_LUNMASK)
	{
	case AEP_F_ASIC :
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "ASIC Command");
		AEP_RequestBuffer		=	&ASIC_CurRequestBuffer;
		AEP_RequestBuffer_mutex	=	&ASIC_RequestBuffer_mutex;
		send_now				=	&ASIC_send_now;
		send_complete			=	&ASIC_send_complete;
		BuffIndex				=	&ASICBuffIndex;
		SendBuffer				=	&SendASICBuffer;
		WRITE_BUFFER_SIZE		=   ASIC_BufferSize;
		break;
	
	case AEP_F_DES :
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "DES Command");
		AEP_RequestBuffer		=	&DES_CurRequestBuffer;
		AEP_RequestBuffer_mutex	=	&DES_RequestBuffer_mutex;
		send_now				=	&DES_send_now;
		send_complete			=	&DES_send_complete;
		BuffIndex				=	&DESBuffIndex;
		SendBuffer				=	&SendDESBuffer;
		WRITE_BUFFER_SIZE		=   DES_BufferSize;
		break;

	case AEP_F_SA :
		AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "AEP_Daemon", AEP_R_OK, "SA Command");
		AEP_RequestBuffer		=	&SA_CurRequestBuffer;
		AEP_RequestBuffer_mutex	=	&SA_RequestBuffer_mutex;
		send_now				=	&SA_send_now;
		send_complete			=	&SA_send_complete;
		BuffIndex				=	&SABuffIndex;
		SendBuffer				=	&SendSABuffer;
		WRITE_BUFFER_SIZE		=   SA_BufferSize;
		break;
	}
 
  /*Determine if any padding is required*/
	padding = message_hdr->datalen & PADDING_MASK;

  /*If padding is not 0 then we need to add some*/
	if (padding)
	{
	  /*The amount needed is 8 - the value of padding.  i.e. if padding was 1 then
		the datalen must be just 1 byte over a quad word boundry, so we must add
		7 bytes of padding (8 - 1).*/
		padding = sizeof(AEP_U64) - padding;
	}
	
	msg_len = message_hdr->datalen + HDR_SIZE + padding;

	/*Make sure the message is not too big for the blocks on the accelerator*/
	if(msg_len > MSG_BLOCK_SIZE)
	{
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDeviceCommand", AEP_R_ARGUMENTS_BAD, "Error : Message too big for Accelerator");
		AEP_free (message_hdr);
			
		return AEP_R_ARGUMENTS_BAD;
	}
	
/*Lock the request buffer mutex.*/
	_LOCK(AEP_RequestBuffer_mutex);
			
  /*If adding this request to the request buffer will overflow the buffer, then
  send it for processing*/
	if (*BuffIndex + msg_len >= WRITE_BUFFER_SIZE)
	{
		if(*BuffIndex != 0)
		{
		  /*The buffer should be sent.  This bool is checked in the request buffer handler thread*/
			*SendBuffer = TRUE;

		  /*Signal the send buffer condition variable*/
			pthread_cond_signal(send_now);

		  /*The request buffer hndlr thread resets the buffer index back to 0 when finished sending,
		  effectively emptying the buffer.  So this comparision will fail when the thread is finished*/
			while((*BuffIndex + msg_len >= WRITE_BUFFER_SIZE) && (*BuffIndex != 0))
			  /*Wait until the request buffer handler thread signals us*/
				pthread_cond_wait(send_complete, AEP_RequestBuffer_mutex);
		}
	}

	BEndToLEndCopy((AEP_U8*)*AEP_RequestBuffer + *BuffIndex,
		(AEP_U8*)message_hdr, HDR_SIZE / sizeof(AEP_U32));
	
  /*Increment BuffIndex the appropriate amount*/
	(*BuffIndex) += HDR_SIZE;

  /*If the message contained any data, read it from the socket directly into the request buffer*/
	if (message_hdr->datalen != 0)
	{
	
		/*Read directly from the socket into the buf*/
	   	rv = AEP_I_ReadFromSocket(sock_fd, 
			(AEP_U8*)*AEP_RequestBuffer + *BuffIndex,
			 message_hdr->datalen);
		
  		if(rv != AEP_R_OK)
		{
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDeviceCommand", rv, "Error : Reading Data from the API");
			(*BuffIndex) -= HDR_SIZE;
			_UNLOCK(AEP_RequestBuffer_mutex);
			
			AEP_free (message_hdr);
			
			return rv;
		}
		
	  /*Increment BuffIndex by the amount of data read.*/
		(*BuffIndex) += message_hdr->datalen;

	  /*If padding is not 0 then we need to add some*/
		if (padding)
		{
		  /*Increment BuffIndex by the amount padding.*/
			(*BuffIndex) += padding;
		}
	}

  /*Used to unlock the request buffer here but sometimes the answer would be back from the
	accelerator before the response item was added to the response table.  So, we must
	hold on to the request buffer mutex until the entry has been added to the response table*/

/*Lock the response table mutex*/
	_LOCK(&response_item_mutex_array[hash_value]);

/*Unlock the request buffer mutex*/
	_UNLOCK(AEP_RequestBuffer_mutex);
		
  /*allocate some memory for the entry that will go into the response table for this tran*/
	pResponse = (AEP_I_RESPONSE_DATA_PTR) AEP_malloc(sizeof(AEP_I_RESPONSE_DATA));
	if(pResponse == NULL_PTR)
	{
		AEP_free (message_hdr);
		_UNLOCK(&response_item_mutex_array[hash_value]);
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDeviceCommand", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");
		return AEP_R_HOST_MEMORY;
	}
	
  /*Prepare to add a request header to the responses hash table*/
	pResponse->transaction_handle = message_hdr->trans_hndl;

  /*Insert the message header into the table entry*/
	pResponse->msg = (AEP_U64_PTR) message_hdr;

  /*No data in response yet*/
	((AEP_MSG_HDR_PTR)pResponse->msg)->datalen = 0;
	
  /*Set up the request's time stamp*/
	gettimeofday(&pResponse->timestamp, NULL);

  /*Add the timeout period*/
	pResponse->timestamp.tv_sec += Request_Timeout;
	
  /*Initialise the response semaphore*/
  	pResponse->response_semaphore = NULL;
	
  /*Depending whether this is a synch or asynch command  and handle appropriately
	If synch wait for reply otherwise just return*/
	if (message_hdr->flags & AEP_F_SYNCMODE)
	{
	  /*Set the Socket descriptor so the response hndlr thread can send the answer back*/
		pResponse->sock_fd = sock_fd;
	  /*Insert the response entry into the hash table*/
		rv = AEP_I_HashPutEntry(AEP_I_response_table.response_table,pResponse->transaction_handle,(AEP_VOID_PTR)pResponse);
		if(rv != AEP_R_OK)
		{
			AEP_free(message_hdr);
			AEP_free(pResponse);
			_UNLOCK(&response_item_mutex_array[hash_value]);
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDeviceCommand", rv, "Error adding Response to Hash Table");

			return rv; 
		}

	  /*Set up the semaphore we'll use to wait for the reply*/
		pResponse->response_semaphore = (sem_t *) AEP_malloc(sizeof(sem_t));
		if (pResponse->response_semaphore == NULL)
		{
		  /*Remove the response entry from the hash table*/
			AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);

			AEP_free (message_hdr);
			AEP_free(pResponse);
			
			_UNLOCK(&response_item_mutex_array[hash_value]);

		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDeviceCommand", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");
			return AEP_R_HOST_MEMORY;
		}

		SEM_INIT(pResponse->response_semaphore, 0, 0);

  	  /*Does not matter that response may have arrived between here and the if
		above as the post will add one to the semaphore and we will not wait at all*/

/*Unlock the response table mutex*/
		_UNLOCK(&response_item_mutex_array[hash_value]);

	  /*Wait for the response to return from the accelerator*/
		SEM_WAIT(pResponse->response_semaphore);

	  /*Uninitialise and free the semaphore used to wait for the reply*/
		sem_destroy(pResponse->response_semaphore);
		AEP_free(pResponse->response_semaphore);

	  /*Make this NULL again, ie no one waiting on semaphore*/
		pResponse->response_semaphore = NULL;
		
		AEP_free(pResponse);

	}
	/*Asynchronous command*/
	else /*if (message_hdr->flags & AEP_F_ASYNCMODE)*/
	{
	  /*No need to use a semaphore*/
		pResponse->response_semaphore = NULL;

	  /*Add the response entry to the hash table*/
		rv = AEP_I_HashPutEntry(AEP_I_response_table.response_table,pResponse->transaction_handle,(AEP_VOID_PTR)pResponse);
		if(rv != AEP_R_OK)
		{
			AEP_free(message_hdr);
			AEP_free(pResponse);
			_UNLOCK(&response_item_mutex_array[hash_value]);
			_UNLOCK(AEP_RequestBuffer_mutex);
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoDeviceCommand", rv, "Error adding Response to Hash Table");


			return rv; 
		}

/*Unlock the request buffer mutex*/
		_UNLOCK(AEP_RequestBuffer_mutex);
		
		/*Unlock the response table mutex*/
		_UNLOCK(&response_item_mutex_array[hash_value]);
	}

	AEP_TRACE(DAEMON, TRACE_DEBUG, FN_OUT, "DoDeviceCommand", AEP_R_OK, "Leaving");

	return rv;
}

/*Function used to process a command that requires a context*/
AEP_RV DoContextCommand(AEP_MSG_HDR_PTR message_hdr, int sock_fd)
{
   /*Pointer to the reply that gets sent back to the app
   Note we need to divide the array size by 8 because it's
   being declared as an array of U64s*/
	AEP_U64 pReply[(HDR_SIZE + 8)>>3];  /*Hdr size of context id size*/
	
   /*Pointer used to read message data from socket in case of an error*/
	AEP_U8_PTR pTrash;

  /*Return value*/
	AEP_U32 rv = AEP_R_OK;

  /*The device command will be added to one of the OutGoing Request Buffers.  Rather than
	write seperate code for each of the buffers we simply check at the start what buffer 
	the request should be added to and then get pointers to the necessary variables and use
	common code to do the work.  The following data structures are used to hold these ptrs*/

  /*Used to point to the RequestBuffer Mutex for this request*/
	pthread_mutex_t *AEP_RequestBuffer_mutex;

  /*Used to point to the condition variables for the sending/waiting to be sent of this buffer*/
	pthread_cond_t *send_now, *send_complete;

  /*Used to point to the offset into the request Buffer for this request*/
	AEP_U32 *BuffIndex;

  /*Used to point to the variable used when deciding if this buffer should go to the accelerator*/
	AEP_U32 *SendBuffer;

  /*Used to point to the Request Buffer for this request*/
	AEP_U64 **AEP_RequestBuffer;

  /*Used to hold the size of the Request buffer associated with this request*/
	AEP_U32 WRITE_BUFFER_SIZE;

  /*Used to determine if and how much padding is necessary*/
	AEP_U32 padding=0;
	
	AEP_U64 ctx_id;
	
	AEP_U32 hash_value;

	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "DoContextCommand", AEP_R_OK, "Entered...");
	
  /*Get the hash value for this transaction*/
	hash_value = AEP_I_HASH(message_hdr->trans_hndl);

  /*Insert log entry*/
    AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "DoContextCommand", AEP_R_OK, "Handling a Command. Type %d", message_hdr->command);

 
  /*Some commands require a context and once one has been assigned, they are processed like any other device commands.
    Others, such as the various init commands do not require a response item to be created.  The context id is passed
	back to the calling app and the request is then simply added to the relevant queue*/
	switch (message_hdr->command)
	{
	/*These commands are assigned a context and then passed to DoDeviceCommand*/
	case AEP_C_AUTH_APP:
/*Lock the Context ID mutex*/
		_LOCK(&AEP_ContextID_mutex);
	
	  /*Make sure there are contexts remaining*/
		if (contextID_count == 0)
		/*Oh NO!  No contexts*/
		{
			_UNLOCK(&AEP_ContextID_mutex);
			
			printf("No ctxs\n");

		  /*If the message contained any data, read it from the socket directly into the request buffer*/
			if (message_hdr->datalen != 0)
			{
			  /*Use pTrash to hold what is read from the socket, then just free it!*/
			  	pTrash = (AEP_U8_PTR) AEP_malloc(message_hdr->datalen);
				
			  /*Malloc failed*/
				if (pTrash == NULL)
				{
					AEP_free(message_hdr);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoContextCommand", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");
					return AEP_R_HOST_MEMORY;
				}
				
			  /*Read the rest of the message from the socket*/
   		 		rv = AEP_I_ReadFromSocket(sock_fd, pTrash, message_hdr->datalen);
				
			  /*Free pTrash ( We don't care about what was in the message, 'cause we aint got no contexts)*/
				AEP_free(pTrash);
				
			  /*The read from socket failed!*/
	  			if(rv != AEP_R_OK)
				{
					AEP_free(message_hdr);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoContextCommand", rv, "Error : Reading Data from the API");
					return rv;
				}
			}
		  /*Mark the reply in err*/
			message_hdr->status = AEP_S_NO_RESOURCES;
		  /*No data in reply*/
			message_hdr->datalen = 0;
		
		  /*Send the completed reply to the app*/
			rv = AEP_I_SendToSocket(sock_fd, (AEP_U8_PTR)message_hdr,
				HDR_SIZE);
								
	  		if(rv != AEP_R_OK)
			{
				AEP_free(message_hdr);
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoContextCommand", rv, "Error : Writing Data to the API");
				return rv;
			}
		
		}
		else
	  /*There are contexts remaining*/
		{
		  /*Assign a context this message, this is a auth_app command so put the context in the app_id*/
			message_hdr->app_id = contextID_array[contextID_count - 1];
			
		  /*We have one less context*/
			contextID_count--;
			
		  /*See if we should request more context ids*/
			if (contextID_count < CONTEXTS_LOW_MARK)
	  		  /*Signal the send buffer condition variable*/
				pthread_cond_signal(&request_contexts);

/*Unlock the COntext ID mutex*/	
			_UNLOCK(&AEP_ContextID_mutex);
		
		  /*We need to convert this from L to B endian now because it will be converted from B to L endian
			on the way to the target.  (Because it is part of the header)*/
			LEndToBEnd((unsigned char*) &message_hdr->app_id, sizeof(AEP_U64) /
				sizeof(AEP_U32));

		  /*Treat this command like any other now*/
			rv = DoDeviceCommand(message_hdr, sock_fd);
			if (rv != AEP_R_OK)
				return rv;
		}
		
	break;
  /*There is no response item set up for this command, but it does not require a context.  The app does not
	wait for confirmation that the UNAUTH command has succeeded*/
	case AEP_C_UNAUTH_APP:
/*Lock the request buffer mutex.*/
		_LOCK(&SA_RequestBuffer_mutex);

	  /*If adding this request to the request buffer will overflow the buffer, then empty it*/
		if (SABuffIndex + HDR_SIZE >= SA_BufferSize)
		{
		  /*The buffer should be sent.  This bool is checked in the request buffer handler thread*/
			SendSABuffer = TRUE;

		  /*Signal the send buffer condition variable*/
			pthread_cond_signal(&SA_send_now);

		  /*The request buffer hndlr thread resets the buffer index back to 0 when finished sending,
	 		 effectively emptying the buffer.  So this comparision will fail when the thread is finished*/
			while(SABuffIndex + HDR_SIZE >= SA_BufferSize)/*BUFFER_SIZE)*/
			  /*Wait until the request buffer handler thread signals us*/
				pthread_cond_wait(&SA_send_complete, &SA_RequestBuffer_mutex);
		}

	  /*Copy the request header into the request buffer*/
	  	memcpy((AEP_U8*)SA_CurRequestBuffer + SABuffIndex, (AEP_U8*)message_hdr, HDR_SIZE);
	
 	 /*Adjust the endianess of the header if necessary*/
 	 	BEndToLEnd((AEP_U8*) (AEP_U8*)SA_CurRequestBuffer + SABuffIndex, HDR_SIZE /
			sizeof(AEP_U32));
	
	  /*Increment BuffIndex the appropriate amount*/
		SABuffIndex += HDR_SIZE;

/*UnLock the request buffer mutex.*/
		_UNLOCK(&SA_RequestBuffer_mutex);
		
		AEP_free(message_hdr);
		
		break;
				
	/*These commands do not require a response item to be set up for them*/
	default:
	
	  /*Determine which request buffer this request is going to be added to*/
		switch(message_hdr->flags & AEP_F_LUNMASK)
		{
		case AEP_F_ASIC :
			AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "DoContextCommand", AEP_R_OK, "ASIC Command");
			AEP_RequestBuffer		=	&ASIC_CurRequestBuffer;
			AEP_RequestBuffer_mutex	=	&ASIC_RequestBuffer_mutex;
			send_now				=	&ASIC_send_now;
			send_complete			=	&ASIC_send_complete;
			BuffIndex				=	&ASICBuffIndex;
			SendBuffer				=	&SendASICBuffer;
			WRITE_BUFFER_SIZE		=   ASIC_BufferSize;
			break;
	
		case AEP_F_DES :
			AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "DoContextCommand", AEP_R_OK, "DES Command");
			AEP_RequestBuffer		=	&DES_CurRequestBuffer;
			AEP_RequestBuffer_mutex	=	&DES_RequestBuffer_mutex;
			send_now				=	&DES_send_now;
			send_complete			=	&DES_send_complete;
			BuffIndex				=	&DESBuffIndex;
			SendBuffer				=	&SendDESBuffer;
			WRITE_BUFFER_SIZE		=   DES_BufferSize;
			break;

		case AEP_F_SA :
			AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "DoContextCommand", AEP_R_OK, "SA Command");
			AEP_RequestBuffer		=	&SA_CurRequestBuffer;
			AEP_RequestBuffer_mutex	=	&SA_RequestBuffer_mutex;
			send_now				=	&SA_send_now;
			send_complete			=	&SA_send_complete;
			BuffIndex				=	&SABuffIndex;
			SendBuffer				=	&SendSABuffer;
			WRITE_BUFFER_SIZE		=   SA_BufferSize;
			break;
		}
	
/*Lock the request buffer mutex.*/
		_LOCK(AEP_RequestBuffer_mutex);

		padding = message_hdr->datalen & PADDING_MASK;

	  /*If padding is not 0 then we need to add some*/
		if (padding)
		{
		  /*The amount needed is 8 - the value of padding.  i.e. if padding was 1 then
			the datalen must be just 1 byte over a quad word boundry, so we must add
			7 bytes of padding (8 - 1).*/
			padding = sizeof(AEP_U64) - padding;
		}
			
	  /*If adding this request to the request buffer will overflow the buffer, then empty it*/
		if (*BuffIndex + message_hdr->datalen + HDR_SIZE + padding >= WRITE_BUFFER_SIZE)
		{
		  /*The buffer should be sent.  This bool is checked in the request buffer handler thread*/
			*SendBuffer = TRUE;

		  /*Signal the send buffer condition variable*/
			pthread_cond_signal(send_now);

		  /*The request buffer hndlr thread resets the buffer index back to 0 when finished sending,
	 		 effectively emptying the buffer.  So this comparision will fail when the thread is finished*/
			while((*BuffIndex + message_hdr->datalen + HDR_SIZE + padding >= WRITE_BUFFER_SIZE) && (*BuffIndex != 0))
			  /*Wait until the request buffer handler thread signals us*/
				pthread_cond_wait(send_complete, AEP_RequestBuffer_mutex);
		}
	
 	 /*Copy the request header into the request buffer*/
  		memcpy((AEP_U8*)*AEP_RequestBuffer + *BuffIndex, (AEP_U8*)message_hdr, HDR_SIZE);
	
 	 /*Adjust the endianess of the header if necessary*/
  		BEndToLEnd((AEP_U8*) *AEP_RequestBuffer + *BuffIndex, HDR_SIZE /
			sizeof(AEP_U32));

	  /*Increment BuffIndex the appropriate amount*/
		(*BuffIndex) += HDR_SIZE;

	  /*If the message contained any data, read it from the socket directly into the request buffer*/
		if (message_hdr->datalen != 0)
		{
   		 	rv = AEP_I_ReadFromSocket(sock_fd,
			(AEP_U8*)*AEP_RequestBuffer + *BuffIndex,
			 message_hdr->datalen);

	  		if(rv != AEP_R_OK)
			{
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoContextCommand", rv, "Error : Reading Data from the API");
				(*BuffIndex) -= HDR_SIZE;
				_UNLOCK(AEP_RequestBuffer_mutex);
				AEP_free(message_hdr);
				return rv;
			}
		
/*Lock the context id mutex*/
			_LOCK(&AEP_ContextID_mutex);
						
		  /*Make sure there are contexts remaining*/
			if (contextID_count == 0)
			/*No contexts*/
			{
				printf("No ctxs2\n");
				_UNLOCK(&AEP_ContextID_mutex);
	
				message_hdr->status = AEP_S_NO_RESOURCES;
				message_hdr->datalen = 0;
				
				(*BuffIndex) -= HDR_SIZE;

				_UNLOCK(AEP_RequestBuffer_mutex);
		
			  /*Send the completed reply to the app*/
				rv = AEP_I_SendToSocket(sock_fd, (AEP_U8_PTR)message_hdr,
					HDR_SIZE);
								
  				if(rv != AEP_R_OK)
				{
					AEP_free(message_hdr);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_DAEMON - DoContextCommand", rv, "Error : Writing Data to the API");
					return rv;
				}
			}
			else
			/*Nice, we have contexts*/
			{
				ctx_id = contextID_array[contextID_count-1];
					
			  /*Dec the context ID count*/
				contextID_count--;

			  /*See if we should request more context ids*/
				if (contextID_count < CONTEXTS_LOW_MARK)
	  		  	  /*Signal the send buffer condition variable*/
					pthread_cond_signal(&request_contexts);

/*Unlock the context ID mutex*/
				_UNLOCK(&AEP_ContextID_mutex);
				
			  /*Insert the Context ID into the message's data, note: the context id is always the first data item*/
				*((AEP_U64*)(*AEP_RequestBuffer + ((*BuffIndex)>>3))) = ctx_id;

			  /*Increment BuffIndex by the amount of data read + any padding.*/
				(*BuffIndex) += message_hdr->datalen + padding;

			  /*Unlock the Rquest Buffer mutex*/
				_UNLOCK(AEP_RequestBuffer_mutex);

			  /*Set up the reply to go back to the app*/

			  /*Set the status of the message to be complete*/
				message_hdr->status = AEP_S_COMPLETE;
		
  			  /*Set the datalen of the message to be the size of an ctx id*/
				message_hdr->datalen = sizeof(AEP_U64);

			  /*Copy the message header into the reply*/
				memcpy(pReply, (AEP_U8_PTR)message_hdr, HDR_SIZE);

			  /*Insert the context id into the reply going back to the app*/
				*(AEP_U64*)(pReply + (HDR_SIZE>>3)) = ctx_id;
						
			  /*Send the reply to the app*/
				rv = AEP_I_SendToSocket(sock_fd, (AEP_U8*)pReply,
				HDR_SIZE + sizeof(AEP_U64));
									
  				if(rv != AEP_R_OK)
				{
					AEP_free(message_hdr);
				    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoContextCommand", rv, "Error : Writing Data to the API");
					return rv;
				}
						
			}

		}
		else
		/*No data with this request therefore it is invalid.  It must at least have 8 bytes of data
		  to accomadate the context id.*/
		{
			/*Mark the reply in err*/
			message_hdr->status = AEP_S_INVALID_ARGS;
			message_hdr->datalen = 0;
				
			/*Take the header out of the request buffer*/
			(*BuffIndex) -= HDR_SIZE;
			
			_UNLOCK(AEP_RequestBuffer_mutex);
		
		  /*Send the reply to the app*/
			rv = AEP_I_SendToSocket(sock_fd, (AEP_U8_PTR)message_hdr,
				HDR_SIZE);
								
  			if(rv != AEP_R_OK)
			{
				AEP_free(message_hdr);
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoContextCommand", rv, "Error : Writing Data to the API");
				return rv;
			}
		}
		AEP_free(message_hdr);		
	break;
	
	}
 
	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_OUT, "DoContextCommand", AEP_R_OK, "Leaving...");
	return rv;
}

/*	DoKeyManCommand
	This function handles key management commands.

	Inputs	:	Request header.  {message_hdr}
				Socket Descriptor of the connection the command is comming over. {sock_fd}
				
	Outputs	:	Returns an AEP_RV, indicating success or failure to the calling function.
*/
AEP_RV DoKeyManCommand(AEP_MSG_HDR_PTR message_hdr, int sock_fd)
{
  /*Temporary buffer*/
	AEP_U64* temp_buf;

  /*Return value*/
	AEP_U32 rv = AEP_R_OK;

  /*Pointer to response hash table element*/
	AEP_I_RESPONSE_DATA_PTR pResponse;
	
	AEP_U32 hash_value;
	
    /*Insert log entry*/
	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "DoKeyManCommand", AEP_R_OK, "Entered...");
     AEP_TRACE(DAEMON, TRACE_DEBUG, FN_NONE, "DoKeyManCommand", AEP_R_OK, "Handling a Command. Type %d", message_hdr->command);

	/*Get the hash value for this transaction*/
	hash_value = AEP_I_HASH(message_hdr->trans_hndl);

  
	
  /*If the message contained any data, read it from the socket directly into the request buffer*/
	if (message_hdr->datalen != 0)
	{
	  /*Get a buffer big enough to hold the msg*/
	  	temp_buf = (AEP_U64*)AEP_malloc(HDR_SIZE + message_hdr->datalen);
	  
	  /*Copy in the msg header*/
	  	memcpy((AEP_U8*)temp_buf, (AEP_U8*)message_hdr, HDR_SIZE);
	
	  /*Read the rest of the msg into the buffer*/
	   	rv = AEP_I_ReadFromSocket(sock_fd, (AEP_U8*)temp_buf + HDR_SIZE, message_hdr->datalen);
			
  		if(rv != AEP_R_OK)
		{
		    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoKeyManCommand", rv, "Error : Reading Data from the API");
			AEP_free(temp_buf);
			AEP_free(message_hdr);
			return rv;
		}
	}
	
	/*Make sure that we have a connection to the key manager*/
	if (keydb_fd == -1)
	{		
		message_hdr->status = AEP_S_NOKEYMANAGER;
		message_hdr->datalen = 0;
		
 	  /*Send bad response to the app*/
		rv = AEP_I_SendToSocket(sock_fd, (AEP_U8_PTR)message_hdr,
				HDR_SIZE);
				
		AEP_free(temp_buf);
		AEP_free(message_hdr);
		
		AEP_TRACE (DAEMON, TRACE_DEBUG, FN_OUT, "DoKeyManCommand", AEP_R_OK, "Leaving...");
		return AEP_R_OK;
	}


  /*allocate some memory for the entry that will go into the response table for this tran*/
	pResponse = (AEP_I_RESPONSE_DATA_PTR) AEP_malloc(sizeof(AEP_I_RESPONSE_DATA));
	if(pResponse == NULL_PTR)
	{
		AEP_free(temp_buf);
		AEP_free(message_hdr);
		
	    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoKeyManCommand", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation Failure");
		return AEP_R_HOST_MEMORY;
	}

  /*Prepare to add a request header to the responses hash table*/
	pResponse->transaction_handle = message_hdr->trans_hndl;

  /*Insert the message header into the table entry*/
	pResponse->msg = (AEP_U64_PTR) message_hdr;

  /*No data in response yet*/
	((AEP_MSG_HDR_PTR)pResponse->msg)->datalen = 0;

  /*Set up the request's time stamp*/
	gettimeofday(&pResponse->timestamp, NULL);

  /*Add the timeout period*/
	pResponse->timestamp.tv_sec += Request_Timeout;
	
/*Lock the key man mutex.*/
	_LOCK(&AEP_KeyMan_mutex);
	
  /*Write the request to the Key Management application*/
  	rv = AEP_I_SendToSocket(keydb_fd, (AEP_U8*)temp_buf, HDR_SIZE +
		((AEP_MSG_HDR_PTR)temp_buf)->datalen);

/*Unlock the keyman mutex*/
	_UNLOCK(&AEP_KeyMan_mutex);

	if (rv != AEP_R_OK)
	{		
	   	AEP_free(pResponse);
		AEP_free(temp_buf);
		
		/*Key manager is no longer listening*/
		keydb_fd = -1;
				
		message_hdr->status = AEP_S_NOKEYMANAGER;
		message_hdr->datalen = 0;
		
 	  /*Send bad response to the app*/
		rv = AEP_I_SendToSocket(sock_fd, (AEP_U8_PTR)message_hdr,
				HDR_SIZE);
				
		AEP_free(message_hdr);
	}
	else
	{
/*Lock the response table mutex*/
		_LOCK(&response_item_mutex_array[hash_value]);
		
	  /*Depending whether this is a synch or asynch command  and handle appropriately
		If synch wait for reply otherwise just return*/
		if (message_hdr->flags & AEP_F_SYNCMODE)
		{
	 	 /*Insert the response entry into the hash table*/
			rv = AEP_I_HashPutEntry(AEP_I_response_table.response_table,pResponse->transaction_handle,(AEP_VOID_PTR)pResponse);
			if(rv != AEP_R_OK)
			{
				AEP_free(temp_buf);
				AEP_free(pResponse->msg);
				AEP_free(pResponse);
				_UNLOCK(&response_item_mutex_array[hash_value]);
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoKeyManCommand", rv, "Error adding Response to Hash Table");

				return rv; 
			}

		  /*Set up the semaphore we'll use to wait for the reply*/
			pResponse->response_semaphore = (sem_t *) AEP_malloc(sizeof(sem_t));
			if (pResponse->response_semaphore == NULL)
			{
				AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
				AEP_free(temp_buf);
				AEP_free(pResponse->msg);
				AEP_free(pResponse);
				_UNLOCK(&response_item_mutex_array[hash_value]);
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoKeyManCommand", AEP_R_HOST_MEMORY, "Error : Host Memory Allocation");
				return AEP_R_HOST_MEMORY;
			}

			SEM_INIT(pResponse->response_semaphore, 0, 0);

	  	  /*Does not matter that response may have arrived between here and the if
			above as the post will add one to the semaphore and we will not wait at all*/
			_UNLOCK(&response_item_mutex_array[hash_value]);
		
			SEM_WAIT(pResponse->response_semaphore);

			_LOCK(&response_item_mutex_array[hash_value]);

		  /*Uninitialise and free the semaphore used to wait for the reply*/
			sem_destroy(pResponse->response_semaphore);
			AEP_free(pResponse->response_semaphore);

		  /*Make this NULL again, ie no one waiting on semaphore*/
			pResponse->response_semaphore = NULL;
  
 		  /*Send the completed reply to the app*/
			rv = AEP_I_SendToSocket(sock_fd, (AEP_U8_PTR)pResponse->msg,
					((AEP_MSG_HDR_PTR)pResponse->msg)->datalen + HDR_SIZE);
			if (rv != AEP_R_OK)
			{			
		 	  /*Remove the response entry from the hash table*/
				AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
				
				AEP_free(temp_buf);
			  	AEP_free(pResponse->msg);
			  	AEP_free(pResponse);

				_UNLOCK(&response_item_mutex_array[hash_value]);

			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoKeyManCommand", rv, "Error : Writing Data to the API");
				return rv;
			}
	
		  /*Remove the response entry from the hash table*/
			rv = AEP_I_HashRemoveEntry(AEP_I_response_table.response_table,pResponse->transaction_handle);
			if(rv != AEP_R_OK)
			{	
				AEP_free(temp_buf);
				AEP_free(pResponse->msg);
				AEP_free(pResponse);
				_UNLOCK(&response_item_mutex_array[hash_value]);
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoKeyManCommand", rv, "Error removing Response from Hash Table");
				return rv; 
			}

		  /*Free the message part of the response item*/
			AEP_free(pResponse->msg);

		  /*Free the mem used to hold the response entry*/
			AEP_free(pResponse);
		}
		/*Asynchronous command*/
		else /*if (message_hdr->flags & AEP_F_ASYNCMODE)*/
		{
		  /*No need to use a semaphore*/
			pResponse->response_semaphore = NULL;

		  /*Add the response entry to the hash table*/
			rv = AEP_I_HashPutEntry(AEP_I_response_table.response_table,pResponse->transaction_handle,(AEP_VOID_PTR)pResponse);
			if(rv != AEP_R_OK)
			{	
				AEP_free(temp_buf);
				AEP_free(message_hdr);
				AEP_free(pResponse);
				_UNLOCK(&response_item_mutex_array[hash_value]);
			    AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "DoKeyManCommand", rv, "Error adding Response to Hash Table");				
				return rv; 
			}
		}  /*end else*/
/*Unlock the response table mutex*/
		_UNLOCK(&response_item_mutex_array[hash_value]);
	}

	/*Free the buffer used to hold the msg data*/
	AEP_free(temp_buf);
	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_OUT, "DoKeyManCommand", AEP_R_OK, "Leaving...");
	return rv;
}


