/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/


/****************************************************************/
/*																*/
/* Filename:	config_hndlr.c									*/
/* Description:	this file contains the implementation of the 	*/
/*				config management functions.					*/
/*																*/
/* Author:		Federica Selva									*/
/* 																*/
/* Product:		AEP1000											*/
/* Date:		03/04/2001										*/
/*																*/
/****************************************************************/


#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <strings.h>

#include "aep_types.h"
#include "aep_i_defs.h"
#include "aep_i_types.h"
#include "aep_daemon_defs.h"

#include "config_defs.h"
#include "config_hndlr.h"
#include "hash.h"
#include "aeptrace.h"


/************************************************************/
/* CONFIG VARIABLES: current value							*/
/*                              							*/
/* Every time you add a new config variable you must also   */
/*    - add its address in p_CurrentConfigValue             */
/*        (attention to the order)                          */
/*    - modify file config.h                                */
/*    - add a macro (#define ) in daemonConfig.h with the   */
/*      item numer   										*/
/*		e.g #define RESPONSE_TABLE_SIZE  0                  */
/************************************************************/

AEP_U32		ResponseTableSize;					/* 0 */
AEP_U32		ASIC_BufferSize; 					/* 1 */ 
AEP_U32		DES_BufferSize;						/* 2 */
AEP_U32		SA_BufferSize;						/* 3 */
AEP_U32		ASIC_WriteTimeout;					/* 4 */
AEP_U32		DES_WriteTimeout;					/* 5 */
AEP_U32		SA_WriteTimeout;					/* 6 */
AEP_U32		ReadBufferSize;						/* 7 */
AEP_U32		Request_Timeout;					/* 8 */
AEP_CHAR	AEP_DeviceName [MAX_STRING_SIZE];	/* 9 */

/***********************************************************/



void * p_CurrentConfigValue [CONFIG_ITEM_NUM] ={
														&ResponseTableSize,	/* 0 */
														&ASIC_BufferSize,	/* 1 */
														&DES_BufferSize,	/* 2 */
														&SA_BufferSize,		/* 3 */
														&ASIC_WriteTimeout,	/* 4 */
														&DES_WriteTimeout,	/* 5 */
														&SA_WriteTimeout,	/* 6 */
														&ReadBufferSize,	/* 7 */
														&Request_Timeout,	/* 8 */
														AEP_DeviceName,     /* 9 */
												};



AEP_U32 AEP_HASH_MASK; 


/***************************************************************/
/***************************************************************/



/****************************************************************
	getConfigItem
	This function parses the daemon config file, searching for a 
	particular config item.

	Inputs	: tempItem	-> is a pointer to a DAEMON_CONFIG_STRUCT 
						   for the item we want to evaluate.
			  itemToGet -> is the item's identifier, corresponding 
						   to the data type field in the struct, 
						   or the entry "item = 2" in the config file. 	
						   
	Outputs	:	if file is not found, or the item is not found in the 
				file a struct with all fields' values set to -1 is 
				returned. If one or more of the item's fields (value, 
				dataSize, dynamic) is not found, a struct is returned 
				with that field's value set to -1.
				Error codes are "or'ed" into the "error" variable which 
				is the function's return value 
****************************************************************/

static AEP_RV getConfigItem(DAEMON_CONFIG_ITEM * tempItem, AEP_U32 itemToGet)
{
	
	FILE *inputFile;
	
	char	line[LINE_LENGTH],
		word[WORD_LENGTH],
		itemWord[WORD_LENGTH],
		valueWord[WORD_LENGTH],
		dataSizeWord[WORD_LENGTH],
		dynamicWord[WORD_LENGTH];
	
	AEP_U32	item, 
		value, 
		dataSize,
		dynamic;
	
	AEP_U64 longValue;
	

	
	int	valueFound,
		dataSizeFound,
		dynamicFound,
		timeToStop,
		error, 
		linecount,
		itemFound,
		lineLength;
	
	AEP_TRACE(DAEMON, TRACE_DEBUG, FN_IN, "getConfigItem", AEP_R_OK, "Entered");
	
	/*These flag indicates when you've found the item you're looking for*/
	itemFound=0;
	timeToStop=0;
	
	/*These values are used to check which fields in the struct we have already parsed
	*When all these are non-zero, we can stop parsing and return the struct
	*(These are needed because fields can occur in any order in the config file (as long as
	*the dataType (aka item) comes first)*/
	valueFound=0;
	dataSizeFound = 0;
	dynamicFound = 0;
	
	error=0;
	linecount=0;
	
	/*Initialise the values in the config item*/
	tempItem->dataType = -1;
	tempItem->value = -1;
	tempItem->dataSize = -1;
	tempItem->longValue = -1;
	tempItem->dynamic = -1;
	
	/*Try to open the configuration file*/
	inputFile=fopen(CONFIG_FILE_NAME, "r");
	if(inputFile==NULL)
	{
		error |= FILE_NOT_FOUND_ERROR;

		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getConfigItem", (AEP_RV)error, "Error : Daemon Configuration File not found");
	}
	
	/* Loop through configuration file */
	/*Parse file until you reach the end, or until you have read the item you want*/
	if(error==0)
	{	
		while((fgets(line, 100, inputFile)!=NULL) && timeToStop==0)
		{
			linecount=0;
			/* Get rid of any space at start or end of this line */
			while(line[linecount]==' ')
				linecount++;
			
			/* Get the length of the line */
			lineLength=0;
			while(line[lineLength]!='\n')
				lineLength++;
			
			
			if(line[linecount]=='\n'){}		/*Ignore blank lines*/
			else if(line[linecount]==';'){}	/*Ignore comment lines*/
			else if(line[linecount]=='['){}	/*Ignore lines containing item names*/
			
			/* Item  - this is the item identifier or data type of the item*/
			else if(line[linecount]=='i')
			{	
				strncpy(word, line + linecount, 4);
				word[4]='\0';
				if (strcmp(word, "item")==0)
				{
					
				/*Check whether you have already read the item you want and this is 
					*the next item. If so, leave the while loop & return the config item*/
					if(itemFound==1)
						break;
					/* Read the item number into item*/ 
					/* e.g. item = 5 */
					
					/* Check that the next part of the string is "=" */
					
					linecount+=4;
					while(line[linecount]==' ')
						linecount++;
					if(line[linecount]=='=')
					{
					/* Copy the rest of the line after the equals sign into
					itemWord and remove any leading spaces.
					This string is the data type specified for the item
						*/
						linecount++;
						while(line[linecount]==' ')
							linecount++;
						strncpy(itemWord, line+linecount, lineLength-3);
						itemWord[lineLength-3]='\0';
						sscanf(itemWord, "%d", &item);
						
						
						/*Check whether this is the item we're looking for - if so, set a flag*/
						if(item==itemToGet)
						{
							tempItem->dataType=item;
							itemFound=1;
						}
						
						
					}/* end if */
					else
					{
						
						error |= CONFIG_FILE_FORMAT_ERROR;
						AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getConfigItem", (AEP_RV)error, "Error in Daemon Configuration File Format");
						
					}/* end else */
				}
			}
			
			/*If you're parsing the item you are looking for, try to read value, min, 
			*max & default. Otherwise skip past these to the next line starting with "item"*/
			else if(itemFound==1)
			{
				/* Value */
				if(line[linecount]=='v')
				{	
					strncpy(word, line, 5);
					word[5]='\0';
					if (strcmp(word, "value")==0)
					{
						/* Read the value and put it in "value" */
						/* e.g. value = 5 */
						
						
						/* Check that the next part of the string is "=" */
						linecount+=5;
						while(line[linecount]==' ')
							linecount++;
						
						if(line[linecount]=='=')
						{
						/* Copy the rest of the line after the equals sign into
						valueWord and remove any leading spaces.
						This string is the value specified for the item
							*/
							linecount++;
							while(line[linecount]==' ')
								linecount++;
							strncpy(valueWord, line+linecount, (lineLength-linecount));
							valueWord[lineLength-linecount]='\0';
							valueFound=1;
							
						}/* end if */
						else
						{
							
							error |= CONFIG_FILE_FORMAT_ERROR;
							AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getConfigItem", (AEP_RV)error, "Error in Daemon Configuration File Format");
								
						}/* end else */
					}
				}
				/* dataSize */
				else if((line[linecount]=='d')&&(line[linecount+1]=='a'))
				{	
					strncpy(word, line+linecount, 8);
					word[8]='\0';
					if (strcmp(word, "dataSize")==0)
					{
						
						/* Read the dataSize value and put it in "dataSize" */
						/* e.g. dataSize = 5 */
						
						/* Check that the next part of the string is "=" */
						linecount+=8;
						while(line[linecount]==' ')
							linecount++;
						
						if(line[linecount]=='=')
						{
						/* Copy the rest of the line after the equals sign into
						dataSizeWord and remove any leading or trailing spaces.
						This string is the dataSize specified for the item (int, long, string)
							*/
							linecount++;
							while(line[linecount]==' ')
								linecount++;
							strncpy(dataSizeWord, line+linecount, lineLength-7);
							dataSizeWord[lineLength-7]='\0';
							sscanf(dataSizeWord, "%d", &dataSize);
							tempItem->dataSize=dataSize;
							dataSizeFound=1;
							
						}/* end if */
						else
						{
							
							error |= CONFIG_FILE_FORMAT_ERROR;
							AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getConfigItem", (AEP_RV)error, "Error in Daemon Configuration File Format");
							
						}/* end else */
					}
				}
				
				/* dynamic */
				else if((line[linecount]=='d')&&(line[linecount+1]=='y'))
				{	
					strncpy(word, line+linecount, 7);
					word[7]='\0';
					
					if (strcmp(word, "dynamic")==0)
					{
						
						/* Read the dynamic value and put it in "dynamic" */
						/* e.g. dynamic = 0 */
						
						/* Check that the next part of the string is "=" */
						linecount+=7;
						while(line[linecount]==' ')
							linecount++;

						if(line[linecount]=='=')
						{
							linecount++;
							while(line[linecount]==' ')
								linecount++;
							strncpy(dynamicWord, line+linecount, lineLength-linecount);
							dynamicWord[lineLength-linecount]='\0';
							
							sscanf(dynamicWord, "%d", &dynamic);
							
							tempItem->dynamic=dynamic;
							dynamicFound=1;
							
						}/* end if */
						else
						{
							
							error |= CONFIG_FILE_FORMAT_ERROR;
							AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getConfigItem", (AEP_RV)error, "Error in Daemon Configuration File Format");
												
						}/* end else */
					}
				}
				
				
			}
			
			/*Check whether all the items we need have been found yet*/
			if((valueFound==1)&&(dataSizeFound==1)&&(dynamicFound==1)){
				
			/*Find out what type of variable it is, and based on this decode 
				the value and put it in the right place*/
				if(tempItem->dataSize == U32_SIZE)
				{
					/*It's an int*/
					sscanf(valueWord, "%u", &value);
					tempItem->value = value;
				}
				else if(tempItem->dataSize == U64_SIZE)
				{
					/*It's a long*/
					sscanf(valueWord, "%llu", &longValue);
					tempItem->longValue = longValue;
				}
				else if(tempItem->dataSize == STRING_SIZE)
				{
					/*It's a string*/
					strncpy(tempItem->stringValue, valueWord, MAX_STRING_SIZE-1);
					tempItem->stringValue[MAX_STRING_SIZE-1] = '\0';
				}
				else
				{
					error |= CONFIG_FILE_FORMAT_ERROR;
					AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getConfigItem", (AEP_RV)error, "Invalid Data Size");
				}
				timeToStop=1;
			}

				
		}/*End While*/
		fclose(inputFile); 
		
	}/*End If no errors*/
	
	
	
	/*Have parsed the entire config file without finding the item*/
	if(itemFound==0)
	{
		error |= ITEM_NOT_FOUND_ERROR;
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getConfigItem", (AEP_RV)error, "Item not found in Configuration File");
	}
	else if((valueFound==0)||(dataSizeFound==0)||(dynamicFound ==0))
	{
		error |= FIELD_NOT_FOUND_ERROR;
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getConfigItem", (AEP_RV)error, "Field missing in Item Entry");
	}
	
	AEP_TRACE(DAEMON, TRACE_DEBUG, FN_OUT, "getConfigItem", AEP_R_OK, "Leaving");
	return error;	
	
}



#if 0
/* THIS FUNCTION IS NOT USED AT THE PRESENT!!!!!!!! */
/****************************************************************
	getAllConfigItems
	This function parses the daemon config file, and returns an 
	array of all items.

	Inputs	: configItems	-> is a pointer to a DAEMON_CONFIG_STRUCT 
							   array that will contain the items.
			  numItemsRead	-> is a pointer to an int that will contain 
							   the number of items read.	
						   
	Outputs	:	If file is not found a struct with all fields' values 
				set to -1 is returned. 
				If one or more of the item's fields (value, dataSize, 
				dynamic) is not found, a struct is returned with that 
				field's value set to -1.
				Error codes are "or'ed" into the "error" variable which 
				is the function's return value
****************************************************************/

static AEP_RV getAllConfigItems(DAEMON_CONFIG_ITEM *configItems, AEP_U32 *numItemsRead)
{
	
	FILE *inputFile;
	
	char	line[LINE_LENGTH],
			word[WORD_LENGTH],
			itemWord[WORD_LENGTH],
			valueWord[WORD_LENGTH],
			dataSizeWord[WORD_LENGTH],
			dynamicWord[WORD_LENGTH];
	
	AEP_U32	item, 
			value, 
			dataSize,
			dynamic;
	
	AEP_U64 longValue;
	DAEMON_CONFIG_ITEM tempItem;
	
	int		valueFound,
			dataSizeFound,
			dynamicFound,
			timeToStop,
			error, 
			linecount,
			itemFound,
			lineLength,
			firstTime,
			itemCount,
			valuesSet;
	
	
	AEP_TRACE(DAEMON, TRACE_DEBUG, FN_IN, "getAllConfigItems", AEP_R_OK, "Entered");

	/*These flag indicates when you've found the item you're looking for*/
	itemFound=0;
	timeToStop=0;
	
	/*Used to indicate the first time through the loop*/
	firstTime =1;
	/*The number of items read so far*/
	itemCount=0;
	
	/*These values are used to check which fields in the struct we have already parsed
	*When all these are non-zero, we can stop parsing and return the struct
	*(These are needed because fields can occur in any order in the config file (as long as
	*the dataType (aka item) comes first)*/
	valueFound=0;
	dataSizeFound = 0;
	dynamicFound = 0;
	valuesSet=0;
	
	error=0;
	linecount=0;
	
	/*Initialise the values in the config item*/
	tempItem.dataType = -1;
	tempItem.value = -1;
	tempItem.dataSize = -1;
	tempItem.longValue = -1;
	tempItem.dynamic = -1;
	
	
	/*Try to open the configuration file*/
	inputFile=fopen(CONFIG_FILE_NAME, "r");
	if(inputFile==NULL)
	{
		error |= FILE_NOT_FOUND_ERROR;
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getAllConfigItems", (AEP_RV)error, "Daemon Configuration File not found");
	}
	
	/* Loop through configuration file */
	/*Parse file until you reach the end*/
	if(error==0)
	{	
		while(fgets(line, 100, inputFile)!=NULL)
		{
		
			
linecount=0;
			/* Get rid of any space at start or end of this line */
			while(line[linecount]==' ')
				linecount++;
			
			/* Get the length of the line */
			lineLength=0;
			while(line[lineLength]!='\n')
				lineLength++;
			
			if(line[linecount]=='\n'){}		/*Ignore blank lines*/
			else if(line[linecount]==';'){}	/*Ignore comment lines*/
			else if(line[linecount]=='['){}	/*Ignore lines containing item names*/
			
			/* Item  - this is the item identifier or data type of the item*/
			/*We use the item identifier to signify the start of a new record, so we have to save the previous 
			  record to the return array if it exists*/
			else if(line[linecount]=='i')
			{	
			
				
				strncpy(word, line + linecount, 4);
				word[4]='\0';
				if (strcmp(word, "item")==0)
				{
					/*If this is not the first item you have read, copy the last one into the array*/
					if(firstTime ==0){
						
						configItems[itemCount].dataType = tempItem.dataType;
						configItems[itemCount].value = tempItem.value;
						configItems[itemCount].longValue = tempItem.longValue;
						strncpy(configItems[itemCount].stringValue, tempItem.stringValue, MAX_STRING_SIZE);
						configItems[itemCount].dynamic = tempItem.dynamic;
						configItems[itemCount].dataSize = tempItem.dataSize;
						
						if((itemFound ==0)||(valueFound==0)||(dataSizeFound==0)||(dynamicFound ==0))
						{
							error |= FIELD_NOT_FOUND_ERROR;
							AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getAllConfigItems", (AEP_RV)error, "Field missing in Item Entry");
						}
						itemCount++;

						/*Reset all the flags that keep track of whether fields in this record have been found*/
						itemFound = 0;
						valueFound =0;
						dataSizeFound = 0;
						dynamicFound = 0;
						/*This flag keep track of whether we've scanned in the values from the strings yet*/
						valuesSet = 0;

					}
					firstTime = 0;

					/* Read the item number into item*/ 
					/* e.g. item = 5 */
					
					/* Check that the next part of the string is "=" */
					
					linecount+=4;
					while(line[linecount]==' ')
						linecount++;
					if(line[linecount]=='=')
					{
					/* Copy the rest of the line after the equals sign into
					itemWord and remove any leading spaces.
					This string is the data type specified for the item
						*/
						linecount++;
						while(line[linecount]==' ')
							linecount++;
						strncpy(itemWord, line+linecount, lineLength-3);
						itemWord[lineLength-3]='\0';
						sscanf(itemWord, "%d", &item);
						tempItem.dataType = item;
						itemFound=1;
						
						
					}/* end if */
					else
					{
						error |= CONFIG_FILE_FORMAT_ERROR;
						AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getAllConfigItems", (AEP_RV)error, "Error in Daemon Configuration File format");
					}/* end else */
				}
			}/*End if item*/ 
			
			/* Value */
			else if(line[linecount]=='v')
			{	
				strncpy(word, line, 5);
				word[5]='\0';
				if (strcmp(word, "value")==0)
				{
					/* Read the value and put it in "value" */
					/* e.g. value = 5 */
					
					
					/* Check that the next part of the string is "=" */
					linecount+=5;
					while(line[linecount]==' ')
						linecount++;
					
					if(line[linecount]=='=')
					{
					/* Copy the rest of the line after the equals sign into
					valueWord and remove any leading spaces.
					This string is the value specified for the item
						*/
						linecount++;
						while(line[linecount]==' ')
							linecount++;
						strncpy(valueWord, line+linecount, (lineLength-linecount));
						valueWord[lineLength-linecount]='\0';
						valueFound=1;
						
					}/* end if */
					else
					{
						
						error |= CONFIG_FILE_FORMAT_ERROR;
						AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getAllConfigItems", (AEP_RV)error, "Error in Daemon Configuration File Format");
					}/* end else */
				}
			}/*end if value*/
			/* dataSize */
			else if((line[linecount]=='d')&&(line[linecount+1]=='a'))
			{	
				strncpy(word, line+linecount, 8);
				word[8]='\0';
				if (strcmp(word, "dataSize")==0)
				{
					
					/* Read the dataSize value and put it in "dataSize" */
					/* e.g. dataSize = 5 */
					
					/* Check that the next part of the string is "=" */
					linecount+=8;
					while(line[linecount]==' ')
						linecount++;
					
					if(line[linecount]=='=')
					{
					/* Copy the rest of the line after the equals sign into
					dataSizeWord and remove any leading or trailing spaces.
					This string is the dataSize specified for the item (int, long, string)
						*/
						linecount++;
						while(line[linecount]==' ')
							linecount++;
						strncpy(dataSizeWord, line+linecount, lineLength-7);
						dataSizeWord[lineLength-7]='\0';
						sscanf(dataSizeWord, "%d", &dataSize);
						tempItem.dataSize=dataSize;
						dataSizeFound=1;
						
					}/* end if */
					else
					{
						error |= CONFIG_FILE_FORMAT_ERROR;
						AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getAllConfigItems", (AEP_RV)error, "Error in Daemon Configuration File Format");
					}/* end else */
				}
			}/*End if dataSize*/
			
			/* dynamic */
			else if((line[linecount]=='d')&&(line[linecount+1]=='y'))
			{	
				strncpy(word, line+linecount, 7);
				word[7]='\0';
				
				if (strcmp(word, "dynamic")==0)
				{
				
					/* Read the dynamic value and put it in "dynamic" */
					/* e.g. dynamic = 0 */
					
					/* Check that the next part of the string is "=" */
					linecount+=7;
					while(line[linecount]==' ')
						linecount++;
					
					if(line[linecount]=='=')
					{
					/* Copy the rest of the line after the equals sign into
					dynamicWord and remove any leading or trailing spaces.
					This string is the dynamic value specified for the item
					*/
						
						linecount++;
						while(line[linecount]==' ')
							linecount++;
						strncpy(dynamicWord, line+linecount, lineLength-linecount);
						dynamicWord[lineLength-linecount]='\0';
						sscanf(dynamicWord, "%d", &dynamic);
						
						tempItem.dynamic=dynamic;
						dynamicFound=1;
						
					}/* end if */
					else
					{
						error |= CONFIG_FILE_FORMAT_ERROR;
						AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getAllConfigItems", (AEP_RV)error, "Error in Daemon Configuration File Format");
					}/* end else */
					
				}
			}/*End if dynamic*/
			
			/*Check whether all the items we need have been found yet*/
			if((valueFound==1)&&(dataSizeFound==1)&&(dynamicFound==1)&&(valuesSet==0)){
				
				

/*Find out what type of variable it is, and based on this decode 
				the value string and put it in the right place*/
				if(tempItem.dataSize == U32_SIZE)
				{
					/*It's an int*/
					sscanf(valueWord, "%u", &value);
					tempItem.value = value;
				}
				else if(tempItem.dataSize == U64_SIZE)
				{
					/*It's a long*/
					sscanf(valueWord, "%llu", &longValue);
					tempItem.longValue = longValue;
				}
				else if(tempItem.dataSize == STRING_SIZE)
				{
					/*It's a string*/
					strncpy(tempItem.stringValue, valueWord, MAX_STRING_SIZE-1);
					tempItem.stringValue[MAX_STRING_SIZE-1] = '\0';
				}
				else
				{
					error |= CONFIG_FILE_FORMAT_ERROR;
					AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getAllConfigItems", (AEP_RV)error, "Invalid Data Size");
				}
				
				valuesSet =1;
				
			
			}/*End if all fields found*/
		}/*End While*/


		
		/*Check why we've left the loop. fgets returns NULL if the eof was reached, or 
		 *if an error occurred. if it was because of an error, we should flag this*/
		 if(ferror(inputFile))
			 error |= ERROR_READING_FILE;


		/*Add the last item to the array*/
		configItems[itemCount].dataType = tempItem.dataType;
		configItems[itemCount].value = tempItem.value;
		configItems[itemCount].longValue = tempItem.longValue;
		strncpy(configItems[itemCount].stringValue, tempItem.stringValue, MAX_STRING_SIZE);
		configItems[itemCount].dynamic = tempItem.dynamic;
		configItems[itemCount].dataSize = tempItem.dataSize;
		itemCount++;

		fclose(inputFile); 
		
	}/*End If no errors*/
	
	
	if((itemFound==0)||(valueFound==0)||(dataSizeFound==0)||(dynamicFound==0))
	{
		error |= FIELD_NOT_FOUND_ERROR;
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "getAllConfigItems", (AEP_RV)error, "Field missing in Item Entry");
	}
	
	*numItemsRead = itemCount;

	AEP_TRACE(DAEMON, TRACE_DEBUG, FN_OUT, "getAllConfigItems", AEP_R_OK, "Leaving");
	return error;	
	
}
#endif /* FUNCTION NOT USED */



/****************************************************************
	ShowError

	Inputs	: rv -> error code
			  dataType -> item identifier
				
	Outputs	: Void
****************************************************************/

static void ShowError (AEP_RV rv, AEP_U32 dataType)
{
	
	switch(rv)
	{
	case ITEM_NOT_FOUND_ERROR:
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ShowError", rv, "dataType not found. dataType : %d", dataType);
		break;
	
	case FILE_NOT_FOUND_ERROR:
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ShowError", rv, "File not found. dataType : %d", dataType);
		break;
		
	case CONFIG_FILE_FORMAT_ERROR:
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ShowError", rv, "Config file format error. dataType : %d", dataType);
		break;

	case FIELD_NOT_FOUND_ERROR:
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ShowError", rv, "Field not found. dataType : %d", dataType);
		break;

	case ERROR_READING_FILE:
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ShowError", rv, "Error reading file. dataType : %d", dataType);
		break;
	
	case NOT_VALID_DATA:
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ShowError", rv, "Not valid data. dataType : %d", dataType);
		break;
	
	default:
		AEP_TRACE (DAEMON, TRACE_ERRORS, FN_NONE, "ShowError", rv, "Unknown error. dataType : %d", dataType);
		break;
	}
}



/****************************************************************
	CheckNewValue
	This function check if the new value to set is valid

	Inputs	: Item -> it holds data about the parameter to set
				
	Outputs	: error code
****************************************************************/

static AEP_RV CheckNewValue (DAEMON_CONFIG_ITEM Item)
{
	int    rv;
	int    i;

	/*
	printf ("dataType : %d\n", Item.dataType);
	printf ("value : %d\n", Item.value);
	*/
	rv = AEP_R_OK;
	
	if (Item.dataType == RESPONSE_TABLE_SIZE)	
	{
		/* Check if the new value to set is a power of 2 */
	
		i = 2;
		
		rv = NOT_VALID_DATA;

		while ((i <= MAX_AEP_RESPONSE_TABLE_SIZE) && (rv != AEP_R_OK))
		{
			if (Item.value != i)
			{
				i = i * 2; 
			}
			else 
			{
				rv = AEP_R_OK;
			}
		}
	}
	return rv;
}



/****************************************************************
	PostAlteration
	This function is called every time there could be a change in 
	a config value, it is needed to maintain the consistency of data

	Inputs	: void
				
	Outputs	: void
****************************************************************/

static void PostAlteration ()
{
	AEP_HASH_MASK = (ResponseTableSize - 1);
}



/****************************************************************
	SetNewValue
	This function set a new value of a parameter

	Inputs	: ConfigItem -> it hold data about the parameter to set
				
	Outputs	: Void
****************************************************************/

static void SetNewValue (DAEMON_CONFIG_ITEM ConfigItem)
{
	AEP_U32		*p_U32;
	AEP_U64		*p_U64;	
	AEP_CHAR	*p_String;


	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_NONE, "SetNewValue", AEP_R_OK, "dataType : %d", ConfigItem.dataType);
	

	if (ConfigItem.dataSize == U32_SIZE)
	{
		p_U32 = p_CurrentConfigValue [ConfigItem.dataType]; 
		/* printf ("    before = %d \n", *p_U32); */
		*p_U32 = ConfigItem.value;
		/* printf ("    after = %d \n", *p_U32); */
		AEP_TRACE (DAEMON, TRACE_DEBUG, FN_NONE, "SetNewValue", AEP_R_OK, "U32_SIZE - New value : %d", *p_U32);
	}
	else if (ConfigItem.dataSize == U64_SIZE)
	{
		p_U64 = p_CurrentConfigValue [ConfigItem.dataType]; 
		/* printf ("    before = %d \n", *p_U64); */
		*p_U64 = ConfigItem.longValue;
		/* printf ("    after = %d \n", *p_U64); */
		AEP_TRACE (DAEMON, TRACE_DEBUG, FN_NONE, "SetNewValue", AEP_R_OK, "U64_SIZE - New value : %d", *p_U64);
	}
	else if (ConfigItem.dataSize == STRING_SIZE)
	{
		p_String = p_CurrentConfigValue [ConfigItem.dataType]; 
		/* printf ("    before = %s \n", p_String); */
		strcpy (p_String, ConfigItem.stringValue);
		/* printf ("    after = %s \n", p_String); */
		AEP_TRACE (DAEMON, TRACE_DEBUG, FN_NONE, "SetNewValue", AEP_R_OK, "STRING_SIZE - New value : %s", p_String);
	}

}


/****************************************************************
	SetConfigValues
	SIGHUP signal handling function

	Inputs	: Void 
				
	Outputs	: Void
****************************************************************/

void SetConfigValues () 
{

	int 			i;
	AEP_RV			rv;
	DAEMON_CONFIG_ITEM	DaemonConfigItem;



	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "SetConfigValues", AEP_R_OK, "Starting (SIGHUP received)...");

	
	for (i = FIRST_CONFIG_ITEM; i <= LAST_CONFIG_ITEM; i++)
	{
		/* This function parses the daemon config file, searching for the i-th 
		   item */
		rv = getConfigItem(&DaemonConfigItem, i);
		
		

		if (rv != AEP_R_OK)
		{
			ShowError (rv, i);
		}
		/* Only if the i-th item is dynamic (dynamic == TRUE) we could change is 
		   value on the fly*/
		else if (DaemonConfigItem.dynamic == TRUE )
		{
			rv = CheckNewValue (DaemonConfigItem);
			if (rv != AEP_R_OK)
			{
				ShowError (rv, i);
			}
			else
			{
				SetNewValue(DaemonConfigItem);
			}
		}
	}

	PostAlteration ();

	/* printf("SIGHUP end \n"); */
	
}



/****************************************************************
	PresetConfigValues
	Set all the value to config at startup 

	Inputs	: Void 
				
	Outputs	: error code
****************************************************************/

AEP_RV PresetConfigValues () 
{
	int 				i;
	AEP_RV				rv;
	DAEMON_CONFIG_ITEM	DaemonConfigItem;
	
	
	/* printf("Preset  \n"); */
	
	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "PresetConfigValues", AEP_R_OK, "Starting...");
	
	rv = AEP_R_OK;
	i  = FIRST_CONFIG_ITEM;

	
	while ((rv == AEP_R_OK) && (i <= LAST_CONFIG_ITEM))
	{

		rv = getConfigItem(&DaemonConfigItem, i);
		
		if (rv != AEP_R_OK)
		{
			ShowError (rv, i);
			return rv;
		}

		rv = CheckNewValue (DaemonConfigItem);
		if (rv != AEP_R_OK)
		{
			ShowError (rv, i);
			return rv;
		}

		SetNewValue(DaemonConfigItem);
		
		i++;
	}
	
	PostAlteration ();

	return rv;
}
