/*****************************************************************************
 *****************************************************************************
 Copyright (c) 2000-2001, AEP Ltd. 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

 1. Redistributions of source code must retain the above copyright notice, 
 this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation 
 and/or other materials provided with the distribution.

 3. Neither the name of AEP Ltd. nor the names of its contributors 
 may be used to endorse or promote products derived from this software 
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
****************************************************************************/

/**********************************************************************
*                                                                     *
* AEP Ltd.                                                            *
*                                                                     *
* This software is supplied under the terms of the license included   *
* above.  All use of this software must be in accordance with the     *
* terms of that license.                                              *
*                                                                     *
**********************************************************************/
/*
	Hash.c
	Diarmuid O'Neill
	13-October-1999

	This file contains the implementation of the hashing functions.
	It includes functions to setup/destroy a hash table, insert/remove an entry, check
	if an entry exists, convert a hash table to a string and hash table iteration fxns.

*/


#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <pthread.h>

#include "aep_i_defs.h"
#include "aep_i_types.h"

#include "aep_daemon_defs.h"

#include "aepmalloc.h"
#include "hash.h"
#include "mutex.h"
#include "aeptrace.h"



extern AEP_U32 AEP_HASH_MASK;

static pthread_mutex_t	AEP_hash_mutex = PTHREAD_MUTEX_INITIALIZER;



/*	AEP_I_HASH:
	Get the index by hashing the key

	Inputs	: Key to be used (For the base of the hash function) {key}
	Outputs	: Index Calculated for the key
*/
AEP_U32 AEP_I_HASH(AEP_U64 key)
{
	return ((key) & (AEP_HASH_MASK));
}



/*	AEP_I_InitHashtable:
	This function sets up a hash table.

	Inputs	: Size of required hash table  {size}
	Outputs	: Pointer to a hash table of the required size.  {ppHashTable}
*/
AEP_RV AEP_I_InitHashtable(
  AEP_I_HASHTABLE_PTR_PTR ppHashTable,  
  AEP_U32 size                     
)
{
  /* Cannot have a hash table of size 0, Apart from the obvious it would result in a div by 0 */
	assert(size!=0);

	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "AEP_I_InitHashtable", AEP_R_OK, "Entered...");

  /*Allocate enough mem for the HASHTABLE structure*/
	*ppHashTable = AEP_calloc(1,sizeof(AEP_I_HASHTABLE));
	if(*ppHashTable == NULL_PTR)
		return AEP_R_HOST_MEMORY;

  /*Set the table size to the required value*/
	(*ppHashTable)->tab_size = size;

  /*Allocate mem for the table*/
	(*ppHashTable)->table = AEP_calloc(size,sizeof(AEP_I_HASH_BUCKET));
	if((*ppHashTable)->table == NULL_PTR)
    {
		AEP_free(*ppHashTable);
		*ppHashTable = NULL_PTR;
		return AEP_R_HOST_MEMORY;
    }

	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_OUT, "AEP_I_InitHashtable", AEP_R_OK, "Leaving...");
	return AEP_R_OK;
}


/*	AEP_I_DestroyHashtable:
	This function destroys and cleans up an existing hash table.

	Inputs	: Pointer to the hash table to be destroyed  {pHashTable}
	Outputs : None
*/
AEP_RV AEP_I_DestroyHashtable(
  AEP_I_HASHTABLE_PTR pHashTable  
)
{
	unsigned int i;
	AEP_I_HASH_BUCKET_PTR next = NULL_PTR;
	AEP_I_HASH_BUCKET_PTR tmp = NULL_PTR;

	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_IN, "AEP_I_DestroyHashtable", AEP_R_OK, "Entered...");
  /*Hash table must exist to be destroyed*/
	if(pHashTable == NULL_PTR) return AEP_R_GENERAL_ERROR;

/*Lock the hash mutex*/
	_LOCK(&AEP_hash_mutex);

  /*Give a yelp if the table contains data*/
	if(pHashTable->entries != 0)
		AEP_TRACE (DAEMON, TRACE_WARNINGS, FN_NONE, "AEP_I_DestroyHashtable", AEP_R_OK, "**** Hashtable contains data! ****");

  /*Go through each index and free any entries present.  If we find one we have to traverse this
	index's list in case there's more.*/
	for(i=0;i<pHashTable->tab_size;i++)
    {
		tmp= pHashTable->table[i];
		while(tmp != NULL_PTR)
		{
			next = tmp->next;
			AEP_free(tmp);
			tmp = next;
		}
	}
  /*Clean up the table itself*/
	AEP_free(pHashTable->table);
	AEP_free(pHashTable);

/*UnLock the hash mutex*/
	_UNLOCK(&AEP_hash_mutex);

	AEP_TRACE (DAEMON, TRACE_DEBUG, FN_OUT, "AEP_I_DestroyHashtable", AEP_R_OK, "Leaving...");

	return AEP_R_OK;
}

/*	AEP_I_HashPutEntry:
	Adds or updates an entry to a hash table

	Inputs	:	Pointer to the destination hash table  {pHashTable}
				Key to be used (For the base of the hash function)  {key}
				The data to be inserted into the tabl  {val}
	
	Outputs	:	None
				
*/
AEP_RV AEP_I_HashPutEntry(
  AEP_I_HASHTABLE_PTR pHashTable,  
  AEP_U64 key,                /* key to be used */
  AEP_VOID_PTR val             /* data to be inserted to the list*/
)
{
	AEP_U32 index; 
	AEP_I_HASH_BUCKET_PTR pBucket = NULL_PTR; 
	AEP_I_HASH_BUCKET_PTR tmpBucket = NULL_PTR;

  /*Hash table must exist*/
	if(pHashTable == NULL_PTR) 
	{
		AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_I_HashPutEntry", AEP_R_GENERAL_ERROR,"NULL Hash Table");
		return AEP_R_GENERAL_ERROR;
	}

  /*Allocate mem for this entry*/
	pBucket = AEP_calloc(1,sizeof(AEP_I_HASH_BUCKET));

	if(pBucket == NULL_PTR)
		return AEP_R_HOST_MEMORY;

  /*Calculate the index for this key*/
	index = AEP_I_HASH(key);

  /* set bucket values */
	pBucket->key = key;
	pBucket->val = val;
	pBucket->index = index;
	pBucket->next = NULL_PTR; /* this should not be necessary 'cause of calloc */

  /*If it's the first entry...*/
	if(pHashTable->table[index] == NULL_PTR)
    {
		pHashTable->table[index]=pBucket;

/*Lock the hash mutex*/
		_LOCK(&AEP_hash_mutex);

		pHashTable->entries++;

/*UnLock the hash mutex*/
		_UNLOCK(&AEP_hash_mutex);
    }
	else
    {
	  /*Must decide whether we are adding or updating and where in the list at the index we
		are to put this entry*/
		tmpBucket= pHashTable->table[index];
		while(TRUE)
		{
			if(key == tmpBucket->key) /* There is an entry with this key, so, update entry */
			{
				tmpBucket->val = pBucket->val;
				AEP_free(pBucket);
				break;
			}
			else if(tmpBucket->next == NULL) /* We've found our place in life */
			{
				tmpBucket->next = pBucket;

/*Lock the hash mutex*/
				_LOCK(&AEP_hash_mutex);

				pHashTable->entries++;

/*UnLock the hash mutex*/
				_UNLOCK(&AEP_hash_mutex);
				break;
			}
			else	/* keep trudging through the list */
				tmpBucket = tmpBucket->next;
		}
    }

	return AEP_R_OK;
}

/*	AEP_I_HashGetEntry:
	Gets a hash entry given the key and the table

	Inputs	:	Pointer to the hash table  {pHashTable}
				Key to be used (For the base of the hash function)  {key}
					
	Outputs	:	Pointer to a Pointer to the data to be extracted from the tabl  {val}
*/
AEP_RV AEP_I_HashGetEntry(
  AEP_I_HASHTABLE_PTR pHashTable,  
  AEP_U64 key,                
  AEP_VOID_PTR_PTR val         
)
{
	AEP_I_HASH_BUCKET_PTR tmpBucket = NULL_PTR;
	AEP_U32 index;

  /* Table must exist */
	if(pHashTable == NULL_PTR) 
	{
		AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_I_HashGetEntry", AEP_R_GENERAL_ERROR,"NULL Hash Table");
		return AEP_R_GENERAL_ERROR;
	}
  /*Perform the hash function to get the index*/
	index = AEP_I_HASH(key);

  /*Is there an entry at this index?*/
	if(pHashTable->table[index] == NULL_PTR)
		return AEP_R_ARGUMENTS_BAD; 
	
	else
    {
		/*Iterate through the list at this index till we find the required entry*/
		tmpBucket= pHashTable->table[index];
		while(TRUE)
		{
			if(key == tmpBucket->key) /* Nice! found our key */
			{
				*val = tmpBucket->val;
				break;
			}
			else if(tmpBucket->next == NULL) /* Whoops, it's not here */
				return AEP_R_ARGUMENTS_BAD; 
			else   /* Try the next item in the list */
				tmpBucket = tmpBucket->next;
		}
    }
	return AEP_R_OK;
}


/*	AEP_I_HashRemoveEntry:
	Removes an entry from a hash table

	Inputs	:	Pointer to the hash table  {pHashTable}
				Key to be used (For the base of the hash function)  {key}
					
*/
AEP_RV AEP_I_HashRemoveEntry(
  AEP_I_HASHTABLE_PTR pHashTable,  
  AEP_U64 key                
)
{
	AEP_U32 index; 
	AEP_I_HASH_BUCKET_PTR currBucket = NULL_PTR;
	AEP_I_HASH_BUCKET_PTR lastBucket = NULL_PTR;

  /*Hash table must exist*/
	if(pHashTable == NULL_PTR)
	{
		AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_I_HashRemoveEntry", AEP_R_GENERAL_ERROR,"NULL Hash Table");
		return AEP_R_GENERAL_ERROR;
	}

  /*Calculate index for this key*/
	index = AEP_I_HASH(key);

  /*Check if entry is in the table*/
	if(pHashTable->table[index] == NULL_PTR)
    {
		return AEP_R_ARGUMENTS_BAD; 
    }
	else
    {
		currBucket = pHashTable->table[index];
		lastBucket = NULL_PTR;

		while(TRUE)
		{
			if(key == currBucket->key) /* we've found it! */
			{
				if(lastBucket == NULL_PTR) 
					/* we're the first entry
					   currBucket->next will be NULL if we're the last entry */
					pHashTable->table[index] = currBucket->next;
				else
					/* close the hole in the list */
					lastBucket->next = currBucket->next;

				AEP_free(currBucket);
/*Lock the hash mutex*/
				_LOCK(&AEP_hash_mutex);

				pHashTable->entries--;

/*UnLock the hash mutex*/
				_UNLOCK(&AEP_hash_mutex);

				break;
			}
			else if(currBucket->next == NULL) /*Couldn't find the entry!*/
			{
				return AEP_R_ARGUMENTS_BAD;
			}
			else			/*Go to the next entry in the list*/
			{
				lastBucket = currBucket;
				currBucket = currBucket->next;
			}
		}
    }
	return AEP_R_OK;
}

/*	AEP_I_HashEntryExists:
	Checks to see if an entry exists for a particular key

	Inputs	:	Pointer to the hash table  {pHashTable}
				Key to be used (For the base of the hash function)  {key}

*/
AEP_RV AEP_I_HashEntryExists(
  AEP_I_HASHTABLE_PTR pHashTable,  /* returned ptr to hash table */
  AEP_U64 key                 /* key of data to be deleted */
)
{
	AEP_U32 index; 
	AEP_I_HASH_BUCKET_PTR currBucket = NULL_PTR;
	AEP_RV rv = AEP_R_OK;

  /*Hash table must exist*/
	if(pHashTable == NULL_PTR) 
	{
		AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_I_HashEntryExists", AEP_R_GENERAL_ERROR,"NULL Hash Table");
		return AEP_R_GENERAL_ERROR;
	}

  /*Get the index by hashing the key*/
	index = AEP_I_HASH(key);

  /*Does this index exist?*/
	if(pHashTable->table[index] == NULL_PTR)
	{
		return AEP_R_ARGUMENTS_BAD; /* index does not exist */
    }
	else
    {
		currBucket = pHashTable->table[index];

	  /*Iterate throught the list at this index until we find/don't find the key.*/
		while(TRUE)
		{	
			if(key == currBucket->key) /* here it is */
			{
				break;
			}
			else if(currBucket->next == NULL) /* key does not exist in the list */
			{
				rv = AEP_R_ARGUMENTS_BAD;
				break;
			}
			else			/* Check the next entry */
			{
				currBucket = currBucket->next;
			}
		}
    }

	return rv;
}


/* handle , space, pointer, newline + new line */
#define AEP_IH_TEXTSIZE 3+1+8+1+1+20

AEP_CHAR_PTR AEP_I_HashtableToString(
  AEP_I_HASHTABLE_PTR pHashTable 
)
{
  AEP_I_HASH_ITERATOR_PTR iter;
  AEP_CHAR_PTR buff = NULL_PTR;
  AEP_CHAR_PTR curr_p = NULL_PTR;
  AEP_U64 key;
  AEP_U32 chars=0;
  AEP_VOID_PTR val = NULL_PTR;

  if(pHashTable == NULL_PTR)
  {
	  AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_I_HashTableToString", AEP_R_GENERAL_ERROR,"NULL Hash Table");
	  return NULL_PTR;
  }

  curr_p = buff = AEP_malloc((AEP_IH_TEXTSIZE*pHashTable->entries)+1);
  if(buff == NULL_PTR)
    return NULL_PTR;

  AEP_I_HashIterateInit(pHashTable, &iter);
  assert( AEP_I_HashIterValid(iter));
  for(; AEP_I_HashIterValid(iter) ; AEP_I_HashIterateInc(iter), curr_p+=chars)
    {
      AEP_I_HashIterateDeRef(iter, &key, &val);

      chars = sprintf(curr_p, "%3lu:0x%p\n", (long int)key, val);
    }

  return buff;
}

AEP_U32 AEP_I_string_hash(
  AEP_CHAR_PTR string
)
{
  AEP_U32 retval =0;
  for(;*string != '\0' ; string++)
    {
      retval += *string;
      retval = retval << 1;
    }
  
  return retval;
}

AEP_RV AEP_I_HashIterateInit(
 AEP_I_HASHTABLE_PTR pHashTable,
 AEP_I_HASH_ITERATOR_PTR * pIterator
)
{
  AEP_U32 i;

  if(pHashTable == NULL_PTR) 
    {
      *pIterator = NULL_PTR;
	  AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_I_HashIterateInit", AEP_R_GENERAL_ERROR,"NULL Hash Table");
      return AEP_R_GENERAL_ERROR;
    }

  /* any elements at all? */
  if(pHashTable->entries == 0)
    {
      *pIterator=NULL_PTR;
      return AEP_R_OK;
    }

  /* get the mem */
  if( (*pIterator = AEP_malloc(sizeof(AEP_I_HASH_ITERATOR))) == NULL_PTR)
    return AEP_R_HOST_MEMORY;

  (*pIterator)->table = pHashTable;

  for(i=0;(i < pHashTable->tab_size) && (pHashTable->table[i] == NULL_PTR);i++);
  if(i < pHashTable->tab_size)
    (*pIterator)->curr_bucket=pHashTable->table[i];
  else
    (*pIterator)=NULL_PTR;

  return AEP_R_OK;
}


/* pIterator is set to NULL_PTR if there is no further element. */
AEP_RV AEP_I_HashIterateInc(
 AEP_I_HASH_ITERATOR_PTR pIterator
)
{
  AEP_U32 i;

  /* paranoia test */
  if(pIterator == NULL_PTR)
  {
	  AEP_TRACE(DAEMON, TRACE_ERRORS, FN_NONE, "AEP_I_HashIterateInc", AEP_R_GENERAL_ERROR,"NULL Hash Table");
	  return AEP_R_GENERAL_ERROR;
  }

  if(pIterator->table == NULL_PTR)
    return AEP_R_GENERAL_ERROR;

  if(pIterator->curr_bucket == NULL_PTR)
    return AEP_R_GENERAL_ERROR;

  if(pIterator->curr_bucket->next != NULL_PTR)
    {
      pIterator->curr_bucket = pIterator->curr_bucket->next;
      return AEP_R_OK;
    }

  /* OK we have to walk along the table to find the next filled bucket */
  for(i = (pIterator->curr_bucket->index)+1; ((i< pIterator->table->tab_size) && 
				 (pIterator->table->table[i] == NULL_PTR) ); i++);
  if(i < pIterator->table->tab_size)
    pIterator->curr_bucket=pIterator->table->table[i];
  else
    {
    pIterator->table=NULL_PTR;
    pIterator->curr_bucket=NULL_PTR;
    }

  return AEP_R_OK;
}

AEP_RV AEP_I_HashIterateDeRef(
 AEP_I_HASH_ITERATOR_PTR pIterator,
 AEP_U64_PTR pKey,
 AEP_VOID_PTR_PTR ppVal
)
{
  /* paranoia test */
  if((pIterator == NULL_PTR) ||
     (pIterator->table == NULL_PTR) ||
     (pIterator->curr_bucket == NULL_PTR))
    return AEP_R_GENERAL_ERROR;

  if(pKey != NULL_PTR)
    *pKey = pIterator->curr_bucket->key;

  if(ppVal != NULL_PTR)
    *ppVal = pIterator->curr_bucket->val;

  return AEP_R_OK;
}

AEP_RV AEP_I_HashIterateDel(
 AEP_I_HASH_ITERATOR_PTR pIterator
)
{
  AEP_I_HASH_ITERATOR_PTR tmp_iter = pIterator;
  AEP_U64 tmp_key;
  AEP_RV rv = AEP_R_OK;

  if((pIterator == NULL_PTR) ||
     (pIterator->table == NULL_PTR) ||
     (pIterator->curr_bucket == NULL_PTR))
    return AEP_R_GENERAL_ERROR;

  rv = AEP_I_HashIterateInc(pIterator);
  if(rv != AEP_R_OK) return rv;

  rv = AEP_I_HashIterateDeRef(tmp_iter, &tmp_key, NULL_PTR);
  if(rv != AEP_R_OK) return rv;

  rv = AEP_I_HashRemoveEntry(pIterator->table, tmp_key);
  
  return rv;
}

/* clear the iterator */
AEP_RV AEP_I_HashIterateDelete(
 AEP_I_HASH_ITERATOR_PTR pIterator
)
{
  AEP_free(pIterator);
  return AEP_R_OK;
}

AEP_BBOOL AEP_I_HashIterValid(
 AEP_I_HASH_ITERATOR_PTR pIterator
)
{
  if((pIterator == NULL) ||
     (pIterator->table == NULL_PTR) ||
     (pIterator->curr_bucket == NULL_PTR))
    return FALSE;

  return TRUE;     
}

/* handle generation functions */
static AEP_U32 AEP_I_tran_id_counter = 1 ; /* must not start with 0 'cause it's AEP_INVALID_HANDLE */

AEP_RV AEP_I_NewTranID(
  AEP_U32_PTR id                 /* handle returned */
)
{
  if(AEP_I_tran_id_counter == I_MAX_HANDLE)
	AEP_I_tran_id_counter = 1;
    
  *id = (AEP_I_tran_id_counter++);

  return AEP_R_OK;
}




